# -*- coding: utf-8 -*
"""Convert ISA-JSOn to MAGE-TAB"""
import logging

from isatools import isajson, magetab


log = logging.getLogger('isatools')


def convert(source_json_fp, out_path):
    """ Converter for ISA-JSON to MAGE-Tab.
    :param source_json_fp: File descriptor of input ISA JSON file
    :param out_path: Output path to write output MAGE-Tab to
    """
    log.info("loading isajson %s", source_json_fp.name)
    ISA = isajson.load(source_json_fp)
    log.info("dumping magetab %s", out_path)
    magetab.dump(ISA, out_path)



# -*- coding: utf-8 -*-
"""Functions for importing from MetabolomicsWorkbench"""
import ftplib
import json
import logging
import os.path
import re
import urllib
from collections import defaultdict
from datetime import date
from urllib.request import urlopen

from bs4 import BeautifulSoup

from isatools import isatab
from isatools.model import (
    Assay,
    Comment,
    Investigation,
    OntologyAnnotation,
    OntologySource,
    Person,
    Protocol,
    ProtocolParameter,
    Publication,
    Study,
    StudyFactor,
)


__author__ = 'proccaserra@gmail.com'

# a method to obtain a block of line between a start and an end marker
# this will be invoked to obtain raw data, metabolite identification,metabolite
# annotation and possible study factors parameters are a filehandle and 2
# strings allowing the specify the section brackets


def getblock(container, start_marker, end_marker):
    try:
        begin = False
        block = []
        lines = str(container).split('\\n')
        # print("lines", str(lines),"\n")
        # lineElements = []
        for line in lines:
            # print("LINE:", line)
            line_elements = line.split('\\t')
            # print("STUFF",line_elements)
            if start_marker in str(line):
                # print("start", line)
                begin = True
            elif end_marker in str(line):
                # print("end", line)
                begin = False

            if begin:
                block.append(line_elements)
        # print("BLOCK",block)
        return block

    except Exception as e:
        logging.exception(e)
        print("Error: in getblock() method, situation not recognized")


# a method of download Metabolomics Workbench archived data from their
# anonymous FTP site
#  input: a valid Metabolomics Workbench study accession number that should
# follow this pattern ^ST\d+[6]

def get_archived_file(mw_study_id):
    success = True
    archive2download = mw_study_id + ".zip"

    try:
        ftp = ftplib.FTP("metabolomicsworkbench.org")
        ftp.login()
        ftp.cwd('Studies')
        ftp.retrlines('LIST')
        ftp.retrbinary("RETR " + archive2download,
                       open(archive2download, 'wb').write)
        ftp.close()
        return success

    except ConnectionRefusedError:
        print("connection refused \n")
    except FileNotFoundError:
        print("file not found on server \n")
    else:
        success = False
        # print("someone broke the internet")
        return success

# a method to create an EBI Metabolights MAF file from Metabolomics Workbench
# REST API over data and metabolites
# input: a valid Metabolomics Workbench study accession number that should
# follow this pattern ^ST\d+[6]


def generate_maf_file(write_dir, mw_study_id, mw_analysis_id):
    try:
        data_url = \
            "http://www.metabolomicsworkbench.org/rest/study/study_id/" \
            + mw_study_id + "/data"
        metabolites_url = \
            "http://www.metabolomicsworkbench.org/rest/study/study_id/" \
            + mw_study_id + "/metabolites"

        with urllib.request.urlopen(data_url) as url:
            try:
                data_response = url.read().decode('utf8')
                data = json.loads(data_response)
            except urllib.error.HTTPError as error:
                data = error.read()

        with urllib.request.urlopen(metabolites_url) as url:
            try:
                metabolites_response = url.read().decode('utf8')
                metabolites = json.loads(metabolites_response)
            except urllib.error.HTTPError as error:
                data = error.read()

        dd = defaultdict(list)
        if len(metabolites) != 0 or len(data) != 0:
            print("generating  maf_file from json API")
            for d in (metabolites, data):
                # you can list as many input dicts as you want here
                for key, value in d.items():
                    # print("KEY: ",key, ": ",value)
                    dd[key].append(value)

            # merging the 2 json feeds and removing duplicated key, since
            # values are always the same
            for k, v in dd.items():
                # print({k: {i: j for x in v for i, j in x.items()}})
                dd[k] = {i: j for x in v for i, j in x.items()}
            try:
                if not isinstance(dd["1"]["DATA"], list):

                    if "other_id" in dd.items():
                        data_rec_header = "metabolite number" + "\t" \
                                          + "metabolite name" \
                                          + "\t" + "metabolite identifier" \
                                          + "\t" + "pubchem identifier" \
                                          + "\t" + "other id" \
                                          + "\t" + "other id type" \
                                          + '\t' + \
                                          ("(" + dd["1"]["units"] + ')\t')\
                            .join(dd["1"]["DATA"].keys()) \
                            + ("(" + dd["1"]["units"] + ")")

                    elif "pubchem_id" in dd.items():
                        data_rec_header = "metabolite number" + "\t" \
                                          + "metabolite name" \
                                          + "\t" + "metabolite identifier" \
                                          + "\t" + "pubchem identifier" \
                                          + '\t' + \
                                          ("(" + dd["1"]["units"] + ')\t')\
                            .join(dd["1"]["DATA"].keys()) \
                            + ("(" + dd["1"]["units"] + ")")

                    else:
                        data_rec_header = "metabolite number" + "\t" \
                                          + "metabolite name" \
                                          + "\t" + "metabolite identifier" \
                                          + '\t' + \
                                          ("(" + dd["1"]["units"] + ')\t')\
                            .join(dd["1"]["DATA"].keys()) \
                            + ("(" + dd["1"]["units"] + ")")
                    with open(write_dir + "/" + mw_study_id + "/data/"
                              + mw_study_id + "_" + mw_analysis_id
                              + "-maf-data-jsonparsing.txt", "w") as fh:
                        # print("writing 'maf file document' to file from
                        # 'generate_maf_file' method:...")
                        fh.writelines(data_rec_header)
                        fh.writelines("\n")

                        for key in dd:
                            # print(dd[key]["analysis_id"])
                            if dd[key]["analysis_id"] == mw_analysis_id:
                                if "other_id" in dd.items():
                                    record_values = \
                                        key + '\t' \
                                        + dd[key]["metabolite_name"] \
                                        + "\t" + dd[key]["metabolite_id"] \
                                        + "\t" + dd[key]["pubchem_id"] \
                                        + "\t" + dd[key]["other_id"] \
                                        + "\t" + dd[key]["other_id_type"]

                                    for value in dd[key]["DATA"].values():
                                        record_values = record_values \
                                            + "\t" + str(value)
                                    fh.writelines(record_values)
                                    fh.writelines("\n")
                                elif "pubchem_id" in dd.items():
                                    record_values = \
                                        key + '\t' \
                                        + dd[key]["metabolite_name"] \
                                        + "\t" \
                                        + dd[key]["metabolite_id"] \
                                        + "\t" + dd[key]["pubchem_id"] \
                                        + "\t" + dd[key]["other_id"] \
                                        + "\t" + dd[key][
                                            "other_id_type"]

                                    for value in dd[key]["DATA"].values():
                                        record_values = \
                                            record_values + "\t" + str(value)
                                    fh.writelines(record_values)
                                    fh.writelines("\n")
                                else:
                                    record_values = \
                                        key + '\t' \
                                        + dd[key]["metabolite_name"] \
                                        + "\t" + dd[key]["metabolite_id"]

                                    for value in dd[key]["DATA"].values():
                                        record_values = \
                                            record_values + "\t" + str(value)
                                    fh.writelines(record_values)
                                    fh.writelines("\n")

                                    # Output resulting json to file
                                    # [Action Dissabled]
                                    # open("output.json", "w").write(
                                    #     json.dumps(dd, sort_keys=True,
                                    #     indent=4, separators=(',', ': '))
                                    # )
                else:
                    print("Dictionary expected, List Found, "
                          "error in MW REST API")
                    # TODO: need feedback from NIH MW to implement a fallback
            except IOError:
                print("input not recognized")

        else:
            if len(data) == 0:
                print("no json feed for data!")
            if len(metabolites) == 0:
                print("no json feed for metabolites!")

    except IOError:
        print("Error: in generate_maf_file() method, situation not recognized")

# a method to obtain the nature of the technology used in the analysis from a
# Metabolomics Workbench Header line the method takes one parameter as
# input: a filehandle the method returns a string holding the ISA
# technology type


def get_assay_type(container):
    assay_type = ""
    try:
        for line in container:
            if "AN:ANALYSIS_TYPE" in str(line):
                if "MS" in str(line):
                    assay_type = "mass spectrometry"
                    # print("assaytype from get_assay_type: ", assay_type)
                elif "NMR" in str(line):
                    assay_type = "nmr spectroscopy"
        return assay_type
    except Exception as e:
        logging.exception(e)
        print("Error: in get_assay_type() method, situation not recognized")


def write_assay(write_dir, technotype, accnum, mw_analysis_nb,
                assayrecords, assay_wf_header):
    try:
        # /Users/Philippe/Documents/git/MW2ISA/
        assayfileoutputpath = write_dir + "/" + accnum + "/"
        if not os.path.exists(assayfileoutputpath):
            os.makedirs(assayfileoutputpath)

        assay_file = open(assayfileoutputpath + "a_" + accnum + "_"
                          + mw_analysis_nb + '.txt', 'w')
        print("writing 'assay information' to file...")
        # DOC: writing header for ISA assay file:
        for this_item in assay_wf_header:
            assay_file.write('"{0}"'.format(this_item))
            assay_file.write('\t')
            # print('"{0}"'.format(this_item))
        assay_file.write("\n")

        # DOC: now writing associated data records:
        if technotype == "mass spectrometry":
            for my_key in assayrecords:
                # print(my_key, str(assayrecords[my_key][0]))
                # DOC: performing insertion of sample identifier in the
                # canonical workflow for ms assay
                assayrecords[my_key][0][0] = my_key
                assayrecords[my_key][0][2] = my_key
                assayrecords[my_key][0][18] = my_key
                assayrecords[my_key][0][19] = my_key + ".mzml"
                assayrecords[my_key][0][23] = my_key

                # writing the record to file
                for this_item in assayrecords[my_key][0]:
                    assay_file.write('"{0}"'.format(this_item))
                    assay_file.write('\t')
                assay_file.write("\n")
            assay_file.write("\n")

        elif technotype == "nmr spectroscopy":
            for my_key in assayrecords:
                # print("mickey: ", my_key)
                # DOC: performing insertion of sample identifier in the
                # canonical workflow for ms assay
                assayrecords[my_key][0][0] = my_key
                assayrecords[my_key][0][2] = my_key
                assayrecords[my_key][0][19] = my_key
                assayrecords[my_key][0][20] = my_key + ".nmrml"
                assayrecords[my_key][0][23] = my_key

                # writing the record to file
                for this_item in assayrecords[my_key][0]:
                    assay_file.write('"{0}"'.format(this_item))
                    assay_file.write('\t')
                assay_file.write("\n")
            assay_file.write("\n")

        assay_file.close()
    except IOError:
        print("Error: in write_assay() method, situation not recognized")
# a method to create Metabolights formated data files which will be referenced
# in the ISA-Tab document the method takes 3 parameters as input: a filehandle,
# a MW identifier for the study, a MW identifier for the analysis the method
# return nothing but creates a raw signal quantification file and a metabolite
# assignment file.


def create_raw_data_files(write_dir, input_techtype, f, input_study_id,
                          input_analysis_id):
    # print("file to download: ", f)
    try:
        # dlurl = urlopen(f)
        # saving a remote file to local drive
        # localcopy = open(studyID + "_" + analysisID + ".txt", 'w+')
        # localcopy.write(str(dlurl.read()))

        # the combination of MW study ID and analysis ID ensure unicity of
        # file name.
        dataoutputdirectory = write_dir + "/" + input_study_id + "/data/"
        if not os.path.exists(dataoutputdirectory):
            os.makedirs(dataoutputdirectory)

        raw_data_file_name = \
            input_study_id + '_' + input_analysis_id + '_raw_data.txt'
        maf_file_name = input_study_id + '_' + input_analysis_id + '_maf.txt'

        if input_techtype == "mass spectrometry":
            dlurl = urlopen(f).read()
            rawblock = getblock(dlurl, "MS_ALL_DATA_START", "MS_ALL_DATA_END")

            if len(rawblock) < 1:
                print("WARNING: no MS raw data reported in MWtab file")
            else:
                print("writing 'ms raw data' to file...")
                with open((dataoutputdirectory + raw_data_file_name), 'w+') \
                        as rawdata:
                    for item in rawblock:
                        print("item: ,", item)
                        rawdata.writelines(
                            "%s\t" % this_element for this_element in item)
                        rawdata.writelines(item)
                        rawdata.writelines("\n")
                        for this_element in item:
                            if "MS_ALL_DATA_START" in this_element:
                                rawdata.writelines(
                                    "datatype:\t%s" % this_element)
                            elif "Bin range" in this_element:
                                rawdata.writelines(
                                    "quantitationtype:   %s" % this_element)
                                rawdata.writelines('\n')
                            else:
                                rawdata.writelines('%s\t' % this_element)
                        rawdata.writelines('\n')

        elif input_techtype == "nmr spectroscopy":
            dlurl = urlopen(f).read()
            rawblock = getblock(
                dlurl, "NMR_BINNED_DATA_START", "NMR_BINNED_DATA_END")
            # rawdata.writelines("%s\n" % item.replace("\t","\\t") for item in
            # rawblock)
            if len(rawblock) < 1:
                print("WARNING: no nmr binned  data reported in MWtab file")
            else:
                print("writing 'nmr binned data' to file...")
                with open((dataoutputdirectory + raw_data_file_name), 'w+') \
                        as rawdata:
                    for item in rawblock:
                        # print(*item, sep='\t')
                        for this_element in item:
                            if "NMR_BINNED" in this_element:
                                rawdata.writelines(
                                    "datatype:\t%s" % this_element)
                            elif "Bin range" in this_element:
                                rawdata.writelines(
                                    "quantitationtype:   %s" % this_element)
                                rawdata.writelines('\n')
                            else:
                                rawdata.writelines('%s\t' % this_element)
                        rawdata.writelines('\n')

        # generate_maf_file(input_study_id)

        if input_techtype == "mass spectrometry":
            dlurl = urlopen(f).read()
            mafblock = getblock(dlurl, "MS_METABOLITE_DATA_START",
                                "MS_METABOLITE_DATA_END")
            mafblock2 = getblock(dlurl, "METABOLITES_START", "METABOLITES_END")
            # print("mafblock2", mafblock2)
            with open((dataoutputdirectory + maf_file_name), 'w+') as mafdata:
                for item in mafblock:
                    mafdata.writelines(
                        "%s\t" % this_element for this_element in item)
                    mafdata.writelines("\n")
                for item in mafblock2:
                    mafdata.writelines(
                        "%s\t" % this_element for this_element in item)
                    mafdata.writelines("\n")
                # mafdata.writelines("%s\n" % item for item in mafblock)
                # mafdata.writelines("%s\n" % item for item in mafblock2)

        elif input_techtype == "nmr spectroscopy":
            dlurl = urlopen(f).read()
            nmr_mafblock = getblock(dlurl, "NMR_METABOLITE_DATA_START",
                                    "NMR_METABOLITE_DATA_END")
            if len(nmr_mafblock) < 1:
                print("WARNING: no nmr metabolite data reported in MWTab")
                # print("nmr_mafblodk:", len(nmr_mafblock))
            else:
                # print("nmr_mafblock", nmr_mafblock)
                # mafdata.writelines("%s\n" % item for item in mafblock)
                with open((dataoutputdirectory + maf_file_name), 'w+') \
                        as mafdata:
                    for item in nmr_mafblock:
                        mafdata.writelines(
                            "%s" % this_element for this_element in item)
    except Exception as e:
        logging.exception(e)
        print("Error in create_raw_data_files() methods, "
              "possibly when trying to write data files")

# a method to create ISA assay tables from an Metabolomics Workbench Study
# Identifier
# the method takes 3 parameters as input: a filehandle, a MW identifier for
# the study, a MW identifier for the analysis
# the method return nothing but creates as many as ISA assay files.


# a method to create an ISA assay table for NMR records
# the method takes a filehandle as input

def create_nmr_assay_records(lol, study_id, analysis_id, fv_records):

    try:
        # print(fv_records)
        # print("getting the nmr MWTab file: ", lol)
        pv_nmr_instrument = "not reported"
        pv_nmr_exprt_type = "not reported"
        pv_nmr_sw_version = "not reported"
        pv_nmr_frequency = "not reported"
        # pv_nmr_field_freq_lock = ""
        # pv_nmr_std_concentration = ""
        pv_nmr_probe = "not reported"
        pv_nmr_solvent = "not reported"
        pv_nmr_tube_size = "not reported"
        pv_nmr_shimming_method = "not reported"
        pv_nmr_pulse_sequence = "not reported"
        # pv_nmr_water_suppression = ""
        # pv_nmr_pulse_width = ""
        # pv_nmr_power_level = ""
        # pv_nmr_receiver_gain = ""
        # pv_nmr_offset_frequency = ""
        pv_nmr_chemical_shift_ref_compound = "not reported"
        # pv_nmr_temp = ""
        pv_nmr_numscan = "not reported"
        # pv_nmr_acquisitiontime = ""
        pv_nmr_acquisition_file = "not reported"
        pv_nmr_relaxationdelay = "not reported"
        # pv_nmr_spectralwidth = ""
        # pv_nmr_acquired_data_points = ""
        pv_nmr_zerofilling = "not reported"
        pv_nmr_apodization = "not reported"
        pv_nmr_baseline_correction_method = "not reported"
        pv_nmr_binned_increment = "not reported"
        pv_nmr_excluded_range = "not reported"
        nmr_rawdata_qt = "NA"
        nmr_maf_qt = "NA"
        # assayrecords = []
        # assay_wf_record = []

        raw_data_file = ""
        assay_wf_header = [
            "Sample Name",
            "Protocol REF",
            "Extract Name",
            "Protocol REF",
            "Parameter Value[nmr instrument]",
            "Parameter Value[nmr type]",
            "Parameter Value[nmr probe]",
            "Parameter Value[spectrometer frequency]",
            "Parameter Value[acquisition file]",
            "Parameter Value[pulse program]",
            "Parameter Value[nmr solvent]",
            "Parameter Value[nmr tube size]",
            "Parameter Value[shimming method]",
            "Parameter Value[relaxation delay]",
            "Parameter Value[number of scans]",
            "Parameter Value[chemical shift reference compound]",
            "Parameter Value[zero filling]",
            "Parameter Value[apodization]",
            "Parameter Value[baseline correction method]",
            "Parameter Value[binned increment]",
            "Parameter Value[binned data exclusion range ]",
            "NMR Assay Name",
            "Free Induction Data File",
            "Protocol REF",
            "Parameter Value[software]",
            "Data Transformation Name",
            "Derived Spectral Data File"]

        input_nmr_file = urlopen(lol).read()
        input_nmr_file = str(input_nmr_file).split('\\n')

        maf_file = str(study_id) + "_" + str(analysis_id) + "_maf_data.txt"

        for this_row in input_nmr_file:
            this_row = this_row.rstrip()
            this_row = str(this_row).split('\\t')
            # print(this_row)

            if "NM:NMR_EXPERIMENT_TYPE" in this_row[0]:
                pv_nmr_exprt_type = this_row[1]

            if "NM:INSTRUMENT_TYPE" in this_row[0]:
                pv_nmr_instrument = this_row[1]

            if "NM:NMR_PROBE" in this_row[0]:
                pv_nmr_probe = this_row[1]

            if "NM:SPECTROMETER_FREQUENCY" in this_row[0]:
                pv_nmr_frequency = this_row[1]

            if "AN:ACQUISITION_PARAMETERS_FILE" in this_row[0]:
                pv_nmr_acquisition_file = this_row[1]

            if "NM:CHEMICAL_SHIFT_REF_CPD" in this_row[0]:
                pv_nmr_chemical_shift_ref_compound = this_row[1]

            if "NM:PULSE_SEQUENCE" in this_row[0]:
                pv_nmr_pulse_sequence = this_row[1]

            if "NM:NUMBER_OF_SCANS" in this_row[0]:
                pv_nmr_numscan = this_row[1]
            #
            # if "AN:ACQUISITION_PARAMETERS_FILE" in row[0]:
            #     pv_nmr_acquisition_file = row[1]
            #     # protocol_files[3] = row[1]

            if "AN:SOFTWARE_VERSION" in this_row[0]:
                pv_nmr_sw_version = this_row[1]

            if "NMR_METABOLITE_DATA:UNITS" in this_row[0] \
                    and len(this_row) > 1:
                nmr_maf_qt = this_row[1]

            if "NMR_ALL_DATA:UNITS" in this_row[0] and len(this_row) > 1:
                nmr_rawdata_qt = this_row[1]

            # if "#ANALYSIS TYPE: NMR" in row[0]:
            #     tt = "nmr spectroscopy"
            # if "NM:NMR_EXPERIMENT_TYPE" in row[0]:
            #     pv_nmr_exprt_type = row[1]
            #
            # if "NM:FIELD_FREQUENCY_LOCK" in row[0]:
            #     pv_nmr_field_freq_lock = row[1]
            # if "NM:STANDARD_CONCENTRATION" in row[0]:
            #     pv_nmr_std_concentration = row[1]
            if "NM:NMR_SOLVENT" in this_row[0]:
                pv_nmr_solvent = this_row[1]
            # if "NM:NMR_TUBE_SIZE" in row[0]:
            #     pv_nmr_tube_size = row[1]
            # if "NM:SHIMMING_METHOD" in row[0]:
            #     pv_nmr_shimming_method = row[1]
            # if "NM:PULSE_SEQUENCE" in row[0]:
            #     pv_nmr_pulse_sequence = row[1]
            # if "NM:WATER_SUPPRESSION" in row[0]:
            #     pv_nmr_water_suppression = row[1]
            # if "NM:PULSE_WIDTH" in row[0]:
            #     pv_nmr_pulse_width = row[1]
            # if "NM:POWER_LEVEL" in row[0]:
            #     pv_nmr_power_level = row[1]
            # if "NM:RECEIVER_GAIN" in row[0]:
            #     pv_nmr_receiver_gain = row[1]
            # if "NM:OFFSET_FREQUENCY" in row[0]:
            #     pv_nmr_offset_frequency = row[1]
            # if "NM:CHEMICAL_SHIFT_REF_CPD " in row[0]:
            #     pv_nmr_chemical_shift_ref_compound = row[1]
            # if "NM:TEMPERATURE" in row[0]:
            #     pv_nmr_temp = row[1]

            # if "NM:DUMMY_SCANS" in this_row[0]:
            #     pv_nmr_dummyscans = this_row[1]
            # if "NM:ACQUISITION_TIME" in this_row[0]:
            #     pv_nmr_acquisitiontime = this_row[1]
            if "NM:RELAXATION_DELAY" in this_row[0]:
                pv_nmr_relaxationdelay = this_row[1]
            # if "NM:SPECTRAL_WIDTH" in this_row[0]:
            #     pv__nmr_spectralwidth = this_row[1]
            # if "NM:NUM_DATA_POINTS_ACQUIRED" in this_row[0]:
            #     pv_nmr_acquired_data_points = this_row[1]
            # if "NM:REAL_DATA_POINTS" in this_row[0]:
            #     pv_nmr_real_data_points = this_row[1]
            if "NM:BINNED_INCREMENT" in this_row[0]:
                pv_nmr_binned_increment = this_row[1]
            if "NM:BINNED_DATA_EXCLUDED_RANGE" in this_row[0]:
                pv_nmr_excluded_range = this_row[1]
            if "NM:ZERO_FILLING" in this_row[0]:
                pv_nmr_zerofilling = this_row[1]
            if "NM:APODIZATION" in this_row[0]:
                pv_nmr_apodization = this_row[1]
            if "NM:BASELINE_CORRECTION_METHOD" in this_row[0]:
                pv_nmr_baseline_correction_method = this_row[1]

            # if "NMR:NMR_EXPERIMENT_TYPE" in row[0]:
            #     pv_nmrexprt_type = row[1]

            # if "MS_METABOLITE_DATA_START" in row[0]:

        # print("assay workflow header:", assay_wf_header)

        assay_wf_backbone_record = [
            "",
            "metabolite extraction protocol",
            "",
            "nuclear magnetic resonance spectroscopy protocol",
            pv_nmr_instrument,
            pv_nmr_exprt_type,
            pv_nmr_probe,
            pv_nmr_frequency,
            pv_nmr_acquisition_file,
            pv_nmr_pulse_sequence,
            pv_nmr_solvent,
            pv_nmr_tube_size,
            pv_nmr_shimming_method,
            pv_nmr_relaxationdelay,
            pv_nmr_numscan,
            pv_nmr_chemical_shift_ref_compound,
            pv_nmr_zerofilling,
            pv_nmr_apodization,
            pv_nmr_baseline_correction_method,
            pv_nmr_binned_increment,
            pv_nmr_excluded_range,
            "",
            raw_data_file,
            "identification protocol",
            pv_nmr_sw_version,
            "",
            maf_file]

        # print("assay workflow backbone: ", assay_wf_backbone_record)
        # assayrecords.append(assay_wf_record)

        # creating assay records
        longrecords = {}
        full_assay_record = assay_wf_backbone_record
        for fv_record in fv_records:
            # print("factors:", fv_record)
            longrecords[fv_record[3]] = [full_assay_record]
            # print("records:", [full_assay_record])
            # print("full assay record: ",fv_record[3], ": ",
            # longrecords[fv_record[3]])

        # print("QTs:", nmr_maf_qt, nmr_rawdata_qt)

        return longrecords, assay_wf_header, nmr_maf_qt, nmr_rawdata_qt

    except BaseException:
        print("Error in create_nmr_assay_records() method, possibly when "
              "trying to write nmr assay data files")


# a method to create an ISA assay table for MS records
# the method takes a filehandle as input

def create_ms_assay_records(
        lol, input_study_id, input_analysis_id, fv_records):

    try:
        # print("checking the factors from create_ms_assay_records: ",
        # fv_records)

        pv_ch_instrument = ""
        pv_ch_column = ""
        pv_ch_flowrate = ""
        pv_ch_inj_temp = ""
        pv_ch_solvent_a = ""
        pv_ch_solvent_b = ""
        pv_ch_type = ""
        pv_ch_file = ""
        pv_ms_type = ""
        pv_ms_ion_mode = ""
        pv_ms_instrument = ""
        pv_ms_acquisitionfile = ""
        pv_ms_analysisfile = ""
        pv_ms_sw_version = ""
        raw_data_file = ""
        ms_rawdata_qt = "NA"
        maf_file = ""
        ms_maf_qt = "NA"

        assayrecords = []
        assay_wf_record = []
        assay_wf_header = ["Sample Name",
                           "Protocol REF",
                           "Extract Name",
                           "Protocol REF",
                           "Parameter Value[chromatography type]",
                           "Parameter Value[chromatography instrument]",
                           "Parameter Value[chromatography column]",
                           "Parameter Value[chromatography setting file]",
                           "Parameter Value[flow rate]",
                           "Parameter Value[injection temperature]",
                           "Parameter Value[solvent a]",
                           "Parameter Value[solvent b]",
                           "Protocol REF",
                           "Parameter Value[mass spectrometry instrument]",
                           "Parameter Value[mass spectrometry type]",
                           "Parameter Value[injection mode]",
                           "Parameter Value[ionization mode]",
                           "Parameter Value[acquisition file]",
                           "MS Assay Name",
                           "Raw Spectral Data File",
                           "Protocol REF",
                           "Parameter Value[analysis file]",
                           "Parameter Value[software]",
                           "Data Transformation Name",
                           "Derived Spectral Data File",
                           "Protocol REF",
                           "Data Transformation Name",
                           "Metabolite Annotation File"
                           ]

        input_ms_file = urlopen(lol).read()
        input_ms_file = str(input_ms_file).split('\\n')

        # print("content of ms file MS?: ", input_ms_file)

        for row_item in input_ms_file:

            row_item = row_item.rstrip()
            row_item = str(row_item).split('\\t')

            # if "AN:ANALYSIS_TYPE" in row_item[0]:
            #     ms_protocol_type = row_item[1].rstrip()
            #
            # if "AN:INSTRUMENT_NAME" in row_item[0]:
            #     pv_instrument = row_item[1]

            if "CH:INSTRUMENT_NAME" in row_item[0]:
                pv_ch_instrument = pv_ch_instrument + " " + row_item[1]

            if "CH:COLUMN_NAME" in row_item[0]:
                pv_ch_column = pv_ch_column + " " + row_item[1]

            if "CH:METHODS_FILENAME" in row_item[0]:
                pv_ch_file = row_item[1]

            if "CH:CHROMATOGRAPHY_TYPE" in row_item[0]:
                pv_ch_type = row_item[1]

            if "CH:INJECTION_TEMP" in row_item[0]:
                pv_ch_inj_temp = row_item[1]

            if "CH:FLOW_RATE" in row_item[0]:
                pv_ch_flowrate = row_item[1]

            if "CH:SOLVENT_A" in row_item[0]:
                pv_ch_solvent_a = pv_ch_solvent_a + " " + row_item[1]

            if "CH:SOLVENT_B" in row_item[0]:
                pv_ch_solvent_b = pv_ch_solvent_a + " " + row_item[1]

            if "MS:INSTRUMENT_NAME" in row_item[0]:
                pv_ms_instrument = row_item[1]

            if "MS:INSTRUMENT_TYPE" in row_item[0]:
                pv_ms_instrument_type = row_item[1]

            if "MS:MS_TYPE" in row_item[0]:
                pv_ms_type = row_item[1]

            if "MS:ION_MODE" in row_item[0]:
                pv_ms_ion_mode = row_item[1]

            if "AN:ACQUISITION_PARAMETERS_FILE" in row_item[0]:
                pv_ms_acquisitionfile = row_item[1]
                # protocol_files[3] = row_item[1]

            if "AN:SOFTWARE_VERSION" in row_item[0]:
                pv_ms_sw_version = row_item[1]

            if "MS_ALL_DATA:UNITS" in row_item[0] and len(row_item) > 1:
                ms_rawdata_qt = row_item[1]

            if "MS_ALL_DATA_START" in row_item[0]:
                raw_data_file = str(
                    input_study_id) + "_" + str(input_analysis_id) \
                    + "_raw_data.txt"

            if "MS_METABOLITE_DATA:UNITS" in row_item[0] and len(row_item) > 1:
                ms_maf_qt = row_item[1]

        maf_file = str(input_study_id) + "_" + str(input_analysis_id) \
            + "_maf_data.txt"
        # print("there", str(input_analysis_id))
        assay_wf_backbone_record = [
            "",
            "metabolite extraction protocol",
            "",
            "chromatography protocol",
            pv_ch_type,
            pv_ch_instrument,
            pv_ch_column,
            pv_ch_file,
            pv_ch_flowrate,
            pv_ch_inj_temp,
            pv_ch_solvent_a,
            pv_ch_solvent_b,
            "mass spectrometry protocol",
            pv_ms_instrument,
            pv_ms_instrument_type,
            pv_ms_type,
            pv_ms_ion_mode,
            pv_ms_acquisitionfile,
            "",
            raw_data_file,
            "identification protocol",
            pv_ms_analysisfile,
            pv_ms_sw_version,
            "",
            raw_data_file,
            "annotation protocol",
            "Metabolite Annotation",
            maf_file,
        ]

        # print("assay workflow: ", str(assay_wf_backbone_record))
        assayrecords.append(assay_wf_record)

        # creating assay records
        longrecords = {}
        full_assay_record = assay_wf_backbone_record

        for fv_record in fv_records:
            # print("factors:", fv_record)
            longrecords[fv_record[3]] = [full_assay_record]
            # print("full assay record: ",fv_record[3], ": ",
            # longrecords[fv_record[3]])

        return longrecords, assay_wf_header, ms_rawdata_qt, ms_maf_qt
    except BaseException:
        print("Error: in create_ms_assay_records() method.")


def get_organism_with_taxid(lol):
    that_species = ""
    that_taxid = ""
    try:
        for this_row in lol:
            if "SU:SUBJECT_SPECIES" in this_row[0]:
                that_species = this_row[1]
            if "SU:TAXONOMY_ID" in this_row[0]:
                that_taxid = this_row[1]
        # print("species & TaxID :", that_species, that_taxid)
        return that_species, that_taxid
    except Exception as e:
        logging.exception(e)
        print("Error in get_organism_with_taxid() method")


def get_fv_records(lol):
    records = []
    factors = {}
    restofrecordheader = []

    for current_row in lol:
        if "SUBJECT_SAMPLE_FACTORS" in str(current_row) \
                and "#" not in str(current_row):
            # print('row from get_fv_records', row)
            if len(current_row) > 2:
                newrecord = []
                if current_row[1] != "-":
                    newrecord.append(current_row[1])
                else:
                    newrecord.append(current_row[2])

                newrecord.append(current_row[2])

                if "|" in current_row[3]:
                    for item in current_row[3].split("|"):
                        factor, value = item.split(":")
                        factor = factor.strip()
                        newrecord.append(value.strip())

                        if factor in factors.keys():
                            factors[factor] += 1
                        else:
                            factors[factor] = 1
                else:
                    elements = current_row[3].split(":")
                    factor = elements[0]
                    if len(elements) > 2:
                        value = elements[1] + elements[2]
                    else:
                        value = elements[1]

                    if factor in factors.keys():
                        factors[factor] += 1
                    else:
                        factors[factor] = 1

                    newrecord.append(value)

                records.append(newrecord)

    for my_key in factors.keys():
        restofrecordheader.append("Factor Value[" + my_key + "]")

    return records, factors, restofrecordheader


def get_mwfile_as_lol(input_url):
    try:
        input_file = urlopen(input_url).read()
        input_file = str(input_file).split('\\n')
        mw_as_lol = []
        for line in input_file:
            lines = line.split('\\t')
            mw_as_lol.append(lines)

        return mw_as_lol
    except IOError:
        print("IOError in get_mwfile_as_lol() method: can not open file or "
              "read data ")

# a method to write an ISA study file
# a


def write_study_file(write_dir, study_acc_num, study_file_header, longrecords):

    try:
        this_study_filename = "s_" + study_acc_num + ".txt"
        # print("study filename: ",this_study_filename)
        # /Users/Philippe/Documents/git/MW2ISA
        studyfilepath = write_dir + "/" + study_acc_num
        if not os.path.exists(studyfilepath):
            os.makedirs(studyfilepath)
        study_file = open((studyfilepath + "/" + this_study_filename), 'w')
        try:
            print("writing 'study sample information' to file...")
            # write study header to file
            # studyfileheader = studyfileheader+factorheader
            for this_element in study_file_header:
                study_file.write('"{0}"'.format(this_element))
                # print('"{0}"'.format(this_element))
                study_file.write('\t')
            study_file.write("\n")

            # writing study records to file
            for each in longrecords:
                # this is to reorder fields following the merge
                each[0], each[1], each[2], each[3] = \
                    each[3], each[0], each[1], each[2]
                each.insert(4, "sample collection protocol")
                each.insert(3, "")
                each[4], each[5] = each[5], each[4]
                each.insert(1, "specimen")

                for item in each:
                    study_file.write("\"" + item + "\"")
                    study_file.write('\t')
                study_file.write('\n')

            study_file.close()

        except IOError:
            print("IOError in write_study_file method(): "
                  "can not write to file.")

        # else:
        #   print("Error in write_study_file method()
        # -something went wrong while trying to write but don't know why!")

    except IOError:
        print("IOError in write_study_file() method: "
              "can not open file or read data ")
    # else:
    #     print("doh, something went wrong but
    # don't know why in write_study_file method()!")


# METHOD: given a Metabolomics Workbench Identifier, download the
# corresponding zip archive via anonymous FTP


def get_raw_data(study_accession_number):
    study_accession_number = str(study_accession_number)
    try:
        ftp_download_url = "ftp://www.metabolomicsworkbench.org/Studies/"\
                           + study_accession_number + ".zip"
        urlopen(url=ftp_download_url)
    except IOError:
        print("IOError in get_raw_data() method: no permission to download "
              "or wrong url")

# METHOD:
#
# a function to iterate over a dictionary of study identifiers matched to a
# technology type: aim is to allow batch
#  processing/download from MW
# dictionary_of_input = {"ST000102": "NMR", "ST000056": "NMR", "ST000282":
# "MS", "ST000367": "MS", "ST000093": "MS",
#                       "ST000159": "MS", "ST000110": "MS", "ST000369": "MS"}
# //////////
# for key in dictionary_of_input.key():
#     page_url = baseurl + dictionary_of_input[key] + "Data&StudyID=" + key
# + "&StudyType=" +
# dictionary_of_input[key] + "&ResultType=1#DataTabs"
#
#     try:
#         process_entry(key, dictionary_of_input[key])
#     except IOError:
#         print()
#     except:
#         print("it does not cut no mustard, isn't it?")
# //////////

# MAIN METHOD:
# "ST000367"
# "ST000093"
# "ST000159"
# "ST000110"
# "ST00036"


def mw2isa_convert(**kwargs):
    options = {
        'studyid': '',
        'outputdir': '',
        'dl_option': '',
        'validate_option': ''}

    conversion_success = True
    try:

        options.update(kwargs)
        print("user options", options)
        studyid = options['studyid']
        outputdir = options['outputdir']
        dl_option = options['dl_option']
        validate_option = options['validate_option']
        # Retrieve DCC MW Webpage for an study entry and obtaining the
        # list of associated 'analysis' (i.a. ISA assays)

        # studyid = input('Enter a study ID name: ')
        # dl_option = input('Download raw data file: yes/no')

        # checking MW study accession number is conform:
        if not re.match(r"(^ST\d{6})", studyid):

            print("this is not a MW accession number, please try again")

        else:
            study_url = "http://www.metabolomicsworkbench.org/rest/study/" \
                        "study_id/" + studyid + "/analysis"

            with urllib.request.urlopen(study_url) as url:
                study_response = url.read().decode('utf8')
                analyses = json.loads(study_response)
                # print("study analysis", analyses)
                if "1" in analyses.keys():
                    print("several analysis")
                    for key in analyses.keys():
                        tt = analyses[key]["analysis_type"]
                        print("analysis_type:", tt)
                else:
                    print("Technology is: ", analyses["analysis_type"])
                    tt = analyses["analysis_type"]

            print("proceeding with MW study identifier: ", studyid,
                  "and technology:", tt)
            # studyid = "ST000367"
            # tt = "MS"
            outputpath = outputdir + "/" + studyid + "/"
            if not os.path.exists(outputpath):
                os.makedirs(outputpath)

            baseurl = "http://www.metabolomicsworkbench.org/data/" \
                      "DRCCMetadata.php?Mode=Study&DataMode="
            page_url = baseurl + tt + "Data&StudyID=" + studyid + \
                "&StudyType=" + tt + "&ResultType=1#DataTabs"
            page = urlopen(page_url).read()
            soup = BeautifulSoup(page, "html.parser")
            AnalysisParamTable = soup.findAll("table", {'class': "datatable2"})

            # analysisid = ""
            # assay_types = []
            # isa_assay_names = []
            # isa_assay_names_with_dlurl = {}

            downLoadURI = "http://www.metabolomicsworkbench.org/data/" \
                          "study_textformat_view.php?STUDY_ID=" + studyid \
                          + "&ANALYSIS_ID="

            study_assays_dict = {"study_id": studyid, "assays": []}
            for table in AnalysisParamTable:
                # ltab = len(table)
                # index = 0
                for index, obj in enumerate(table):
                    if "Analysis ID:" in str(obj):
                        tds = obj.find_all('td')
                        analysisid = tds[1].text
                        # print(index,"analysis ID", table.tr.next(),
                        # analysisid)
                        if "MS" in str(table.tr.next()):
                            tt = "mass spectrometry"
                            study_assays_dict["assays"].append(
                                {"analysis_id": analysisid, "techtype": tt})
                        elif "NMR" in str(table.tr.next()):
                            tt = "nmr spectroscopy"
                            study_assays_dict["assays"].append(
                                {"analysis_id": analysisid, "techtype": tt})

            # DOC: This print statement shows we are getting all the possible
            # analysis for a given study
            # print("all analysis", study_assays_dict["assays"])

            # Initialization of a range of variables mapping to ISA
            # investigation and study files
            study_title = ""
            study_desc = ""
            study_design = ""
            # study_subdate = ""
            # study_releasedate = ""
            # study_accnum = ""
            study_funder = ""

            # ms_protocol = ""
            treat_protocol = ""
            # plant_treatment_protocol = ""
            plant_plot_design = ""
            plant_light_period = ""
            # plant_humidity = ""
            plant_temperature = ""
            plant_nutriregime = ""
            plant_harvest_method = ""
            quenching_method = ""

            collect_protocol = ""
            sampleprep_protocol = ""
            chromatography_protocol = ""
            mass_spec_protocol = ""
            nmrspc_protocol = ""

            chromatography_protocol_params = []
            ms_protocol_params = []
            nmr_protocol_params = []
            # ms_analysis_params = []
            # nmr_analysis_params = []
            ident_protocol_params = []

            # mt = "metabolite profiling"

            # protocol_names = ["", "", "", "", "", "", "", "", ""]
            # protocol_types = ["sample collection", "treatment",
            #                   "metabolite extraction", "sample preparation",
            #                   "chromatography",
            #                   "mass spectrometry", "nmr spectroscopy",
            #                   "identification", "annotation"]
            protocol_descriptions = ["", "", "", "", "", "", "", "", ""]
            protocol_parameters = {"sample collection": [""],
                                   "treatment": [""],
                                   "metabolite extraction": [""],
                                   "sample preparation": [""],
                                   "chromatography": [""],
                                   "mass spectrometry": [""],
                                   "nmr spectroscopy": [""],
                                   "identification": [""],
                                   "annotation": [""]}
            protocol_files = ["", "", "", "", "", "", "", "", ""]

            study_person_firstname = ""
            study_person_lastname = ""
            study_person_email = ""
            study_person_affiliation = ""
            study_person_address = ""
            # study_person_phone = ""
            study_person_fax = ""
            study_createdon = ""
            study_submittedon = ""
            # study_convertedon = ""
            # study_num_group = ""
            # study_total_subj = ""
            # study_version = ""

            # material_type = ""
            # sample_type = ""
            # organism = ""

            # study_factor_records = []
            # study_factors = {}
            # fv_record_header = []

            basestudysamplerecordheader = [
                "Source Name", "Material Type", "Characteristics[Organism]",
                "Term Accession Number",
                "Term Source REF", "Protocol REF", "Sample Name"]
            # //// END OF VARIABLE DECLARATION

            # Getting a MWTab file using the first analysis (There are as many
            # MWTab file as there are analysis
            # and they all share common section and information about samples
            #  and protocols so we get the first file
            # to prime.
            analysisid = study_assays_dict["assays"][0]["analysis_id"]
            downLoadURI = downLoadURI + analysisid + "&MODE=d"
            # Going over the firt MWtab and loading it as an array of lines
            thisFileContent = get_mwfile_as_lol(downLoadURI)

            # Generating the ISA Study Sample Table stub from MW Tab file
            # Factor section:
            study_factor_records, study_factors, fv_record_header = \
                get_fv_records(thisFileContent)
            # Getting Sample Organism information:
            species, taxonid = get_organism_with_taxid(thisFileContent)
            # Inserting the taxonomic information in the ISA Study
            # Sample Table stub.
            for each_record in study_factor_records:
                each_record.insert(0, species)
                each_record.insert(1, taxonid)

            # Building the Investigation Object and its elements:
            investigation = Investigation(identifier=studyid)
            investigation.comments.append(Comment(
                name="Primary Database", value="NIH Metabolomics Workbench"))
            investigation.comments.append(Comment(
                name="conversion date", value=str(date.today())))
            investigation.comments.append(Comment(
                name="conversion software", value="MW2ISA version 1.0"))
            investigation.comments.append(Comment(
                name="conversion performer email",
                value="philippe.rocca-serra@oerc.ox.ac.uk"))

            investigation.studies.append(Study(identifier=studyid))
            study1 = investigation.studies[0]
            study1.comments = list()

            # Scannning the MWTab file for common information and setting
            # values to variables
            for row in thisFileContent:

                if str(row[0]).startswith('VERSIO'):
                    # study_version = row[1]
                    study1.comments.append(Comment(
                        name="Version", value=row[1]))

                if row[0].find('CREATED_ON') != -1:
                    study_createdon = row[1]
                    study1.comments.append(Comment(
                        name="MW creation date", value=row[1]))

                if row[0].find('ST:SUBMIT_DATE') != -1:
                    study_submittedon = row[1]
                    study1.comments.append(Comment(
                        name="MW submission date", value=row[1]))

                if row[0].find('ST:NUM_GROUPS') != -1:
                    # study_num_group = row[1]
                    study1.comments.append(Comment(
                        name="number of study groups", value=row[1]))

                if row[0].find('ST:TOTAL_SUBJECTS') != -1:
                    # study_total_subj = row[1]
                    study1.comments.append(Comment(
                        name="total number of subjects", value=row[1]))

                if row[0].find('ST:STUDY_TITLE') != -1:
                    study_title = study_title + row[1]

                if row[0].find('ST:STUDY_TYPE') != -1:
                    study_design = row[1]

                if row[0].find('ST:STUDY_SUMMARY') != -1:
                    study_desc = study_desc + " " + row[1]

                if row[0].find('ST:LAST_NAME') != -1:
                    study_person_lastname = row[1]

                if row[0].find('ST:FIRST_NAME') != -1:
                    study_person_firstname = row[1]

                if row[0].find('ST:EMAIL') != -1:
                    study_person_email = row[1]

                # if row[0].find('ST:PHONE') != -1:
                #     study_person_phone = row[1]

                # if row[0].find('ST:SUBMIT_DATE') != -1:
                #     study_subdate = row[1]

                if row[0].find('ST:INSTITUTE') != -1:
                    study_person_affiliation = row[1]

                if row[0].find('ST:DEPARTMENT') != -1:
                    study_person_affiliation = \
                        study_person_affiliation + ", " + row[1]

                if row[0].find('ST:LABORATORY') != -1:
                    study_person_affiliation = \
                        study_person_affiliation + ", " + row[1]

                if row[0].find('PR:ADDRESS') != -1:
                    study_person_address = study_person_address + ", " + row[1]

                if row[0].find('PR:FUNDING_SOURCE') != -1:
                    study_funder = study_funder + " " + row[1]

                if row[0].find('CO:COLLECTION_SUMMARY') != -1:
                    collect_protocol = collect_protocol + " " + row[1]

                if row[0].find('TR:TREATMENT_SUMMARY') != -1:
                    treat_protocol = treat_protocol + " " + row[1]

                # GETTING CHROMATOGRAPHY PARAMETERS

                if row[0].find('CH:CHROMATOGRAPHY_SUMMARY') != -1:
                    chromatography_protocol = \
                        chromatography_protocol + " " + row[1]

                if row[0].find('CH:INSTRUMENT_NAME') != -1:
                    # chromatography_protocol =
                    # chromatography_protocol + " " + row[1]
                    chromatography_param_oa = OntologyAnnotation(
                        term="chromatography instrument")
                    chromatography_param = ProtocolParameter(
                        parameter_name=chromatography_param_oa)
                    chromatography_protocol_params.append(chromatography_param)

                if row[0].find('CH:CHROMATOGRAPHY_TYPE') != -1:
                    # protocol_parameters["chromatography"].append(
                    # "chromatography type: "  + row[1])
                    chromatography_param_oa = OntologyAnnotation(
                        term="chromatography type")
                    chromatography_param = ProtocolParameter(
                        parameter_name=chromatography_param_oa)
                    chromatography_protocol_params.append(
                        chromatography_param)

                if row[0].find('CH:COLUMN_NAME') != -1:
                    # protocol_parameters["chromatography"].append(
                    # "chromatography column name: "  + row[1])
                    chromatography_param_oa = OntologyAnnotation(
                        term="chromatography column")
                    chromatography_param = ProtocolParameter(
                        parameter_name=chromatography_param_oa)
                    chromatography_protocol_params.append(
                        chromatography_param)

                if row[0].find('CH:COLUMN_TEMPERATURE') != -1:
                    # protocol_parameters["chromatography"].append(
                    # "chromatography temperature: "  + row[1])
                    chromatography_param_oa = OntologyAnnotation(
                        term="chromatography column temperature")
                    chromatography_param = ProtocolParameter(
                        parameter_name=chromatography_param_oa)
                    chromatography_protocol_params.append(chromatography_param)

                if row[0].find('CH:TRANSFERLINE_TEMPERATURE') != -1:
                    # protocol_parameters["chromatography"].append(
                    # "transferline temperature: "  + row[1])
                    chromatography_param_oa = OntologyAnnotation(
                        term="transferline temperature")
                    chromatography_param = ProtocolParameter(
                        parameter_name=chromatography_param_oa)
                    chromatography_protocol_params.append(chromatography_param)

                if row[0].find('CH:WASHING_BUFFER') != -1:
                    # protocol_parameters["chromatography"].append(
                    # "washing buffer: "  + row[1])
                    chromatography_param_oa = OntologyAnnotation(
                        term="washing buffer")
                    chromatography_param = ProtocolParameter(
                        parameter_name=chromatography_param_oa)
                    chromatography_protocol_params.append(chromatography_param)

                if row[0].find('CH:SAMPLE_LOOP_SIZE') != -1:
                    # protocol_parameters["chromatography"].append(
                    # "sample loop size: "  + row[1])
                    chromatography_param_oa = OntologyAnnotation(
                        term="sample loop size buffer")
                    chromatography_param = ProtocolParameter(
                        parameter_name=chromatography_param_oa)
                    chromatography_protocol_params.append(chromatography_param)

                if row[0].find('CH:OVEN_TEMPERATURE') != -1:
                    # protocol_parameters["chromatography"].append
                    # ("oven temperature: "  + row[1])
                    chromatography_param_oa = OntologyAnnotation(
                        term="oven temperature")
                    chromatography_param = ProtocolParameter(
                        parameter_name=chromatography_param_oa)
                    chromatography_protocol_params.append(chromatography_param)

                if row[0].find('CH:FLOW_RATE') != -1:
                    # protocol_parameters["chromatography"].append(
                    # "flow rate: "  + row[1])
                    chromatography_param_oa = OntologyAnnotation(
                        term="flow rate")
                    chromatography_param = ProtocolParameter(
                        parameter_name=chromatography_param_oa)
                    chromatography_protocol_params.append(chromatography_param)

                if row[0].find('CH:SOLVENT_A') != -1:
                    # protocol_parameters["chromatography"].append(
                    # "flow rate: "  + row[1])
                    chromatography_param_oa = OntologyAnnotation(
                        term="solvent a")
                    chromatography_param = ProtocolParameter(
                        parameter_name=chromatography_param_oa)
                    chromatography_protocol_params.append(chromatography_param)

                if row[0].find('CH:SOLVENT_B') != -1:
                    # protocol_parameters["chromatography"].append(
                    # "flow rate: "  + row[1])
                    chromatography_param_oa = OntologyAnnotation(
                        term="solvent b")
                    chromatography_param = ProtocolParameter(
                        parameter_name=chromatography_param_oa)
                    chromatography_protocol_params.append(chromatography_param)

                if row[0].find('CH:INJECTION_TEMPERATURE') != -1:
                    # protocol_parameters["chromatography"].append(
                    # "oven temperature: "  + row[1])
                    chromatography_param_oa = OntologyAnnotation(
                        term="injection temperature")
                    chromatography_param = ProtocolParameter(
                        parameter_name=chromatography_param_oa)
                    chromatography_protocol_params.append(chromatography_param)

                # GETTING MS PARAMETERS
                if row[0].find("MS:MS_COMMENTS") != -1:
                    mass_spec_protocol = mass_spec_protocol + " " + row[1]

                if row[0].find('MS:INSTRUMENT_NAME') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "mass spectrometry instrument: "  + row[1])
                    ms_param_oa = OntologyAnnotation(
                        term="mass spectrometry instrument")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:INSTRUMENT_TYPE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "mass spectrometry type: "  + row[1])
                    ms_param_oa = OntologyAnnotation(
                        term="mass spectrometry instrument type")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:MS_TYPE') != -1:
                    # protocol_parameters["mass spectrometry"].append
                    # ("inlet type: "  + row[1])
                    ms_param_oa = OntologyAnnotation(
                        term="mass spectrometry type")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:ION_SOURCE_TEMPERATURE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "ion source temperature: "  + row[1])
                    ms_param_oa = OntologyAnnotation(
                        term="ion source temperature")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:ION_MODE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "ion energy: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="ion mode")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:ION_ENERGY') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "ion energy: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="ion energy")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:ION_SPRAY_VOLTAGE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "ion spray voltage: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="ion spray voltage")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:MASS_ACCURACY') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "mass accuracy: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="mass accuracy")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:SOURCE_TEMPERATURE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "source temperature: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="source temperature")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:SCAN_RANGE_MOVERZ') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "scan range (m/z): "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="scan range (m/z)")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:SCANNING_CYCLE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "scanning cycle: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="scan range (m/z)")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:SCANNING') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "scanning: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="scanning")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:SCANNING_RANGE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "scanning range: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="canning range")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:CAPILLARY_TEMPERATURE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "capillary temperature: "  + row[1])
                    ms_param_oa = OntologyAnnotation(
                        term="capillary temperature")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:SKIMMER_VOLTAGE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "skimmer voltage: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="skimmer voltage")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:TUBE_LENS_VOLTAGE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "tube lens voltage: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="tube lens voltage")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:CAPILLARY_VOLTAGE') != -1:
                    protocol_parameters["mass spectrometry"].append(
                        "capillary voltage: " + row[1])
                    ms_param_oa = OntologyAnnotation(term="capillary voltage")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:COLLISION_ENERGY') != -1:
                    protocol_parameters["mass spectrometry"].append(
                        "collision energy: " + row[1])
                    ms_param_oa = OntologyAnnotation(term="collision energy")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param_oa)

                if row[0].find('MS:COLLISION_GAS') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "collision gas: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="collision gas")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:DRY_GAS_FLOW') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "dry gas flow: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="dry flow gas")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:DRY_GAS_TEMPERATURE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "dry gas: " + row[1])
                    ms_param_oa = OntologyAnnotation(term="dry gas")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:FRAGMENT_VOLTAGE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "fragment voltage: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="fragment voltage")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:FRAGMENTATION_METHOD') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "fragmentation method: "  + row[1])
                    ms_param_oa = OntologyAnnotation(
                        term="fragmentation method")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:GAS_PRESSURE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "gas pressure: " + row[1])
                    ms_param_oa = OntologyAnnotation(term="gas pressure")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:HELIUM_FLOW') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "helium flow: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="helium flow")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:DESOLVATION_GAS_FLOW') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "desolvation gas flow: "  + row[1])
                    ms_param_oa = OntologyAnnotation(
                        term="desolvation gas flow")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:DESOLVATION_TEMPERATURE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "desolvation temperature: "  + row[1])
                    ms_param_oa = OntologyAnnotation(
                        term="desolvation temperature")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:ACTIVATION_PARAMETER') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "activation parameter: "  + row[1])
                    ms_param_oa = OntologyAnnotation(
                        term="activation parameter")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:ACTIVATION_TIME') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "activation time: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="activation time")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:ATOM_GUN_CURRENT') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "atom gun current: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="atom gum current")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:AUTOMATIC_GAIN_CONTROL') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "automatic gain control: "  + row[1])
                    ms_param_oa = OntologyAnnotation(
                        term="automatic gain control")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:BOMBARDMENT') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "bombardment: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="bombardment")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:LASER') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "laser: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="laser")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:MATRIX') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "matrix: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="amtrix")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:NEBULIZER') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "nebulizer: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="nebulizer")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:OCTPOLE_VOLTAGE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "octpole voltage: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="octpole voltage")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:IT_SIDE_OCTPOLES_BIAS_VOLTAGE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "IT side octpole bias voltage: "  + row[1])
                    ms_param_oa = OntologyAnnotation(
                        term="IT side octpole bias voltage")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:CDL_SIDE_OCTPOLES_BIAS_VOLTAGE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "CDL side octpole bias voltage: "  + row[1])
                    ms_param_oa = OntologyAnnotation(
                        term="CDL side octpole bias voltage")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:CDL_TEMPERATURE') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "CDL temperature: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="CDL temperature")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:PROBE_TIP') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "probe tip: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="probe tip")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:RESOLUTION_SETTING') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "resolution setting: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="resolution setting")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                if row[0].find('MS:SAMPLE_DRIPPING') != -1:
                    # protocol_parameters["mass spectrometry"].append(
                    # "sample dripping: "  + row[1])
                    ms_param_oa = OntologyAnnotation(term="sample dripping")
                    ms_param = ProtocolParameter(parameter_name=ms_param_oa)
                    ms_protocol_params.append(ms_param)

                # GETTING NMR PARAMETERS
                if row[0].find('NMR:NMR_SUMMARY') != -1:
                    nmrspc_protocol = nmrspc_protocol + " " + row[1]

                if row[0].find('NMR:INSTRUMENT_NAME') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "nmr instrument: "  + row[1])
                    nmr_param_oa = OntologyAnnotation(term="nmr instrument")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:EXPERIMENT_TYPE') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "nmr experiment type: " + row[1])
                    nmr_param_oa = OntologyAnnotation(
                        term="nmr experiment type")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:RANDOMIZATION_ORDER') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "randomization order: " + row[1])
                    nmr_param_oa = OntologyAnnotation(
                        term="randomization order")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:FIELD_FREQUENCY_LOCK') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "field frequency lock: " + row[1])
                    nmr_param_oa = OntologyAnnotation(
                        term="field frequency lock")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:STANDARD_CONCENTRATION') != -1:
                    # protocol_parameters["nmr spectroscopy"].append
                    # ("standard concentration: " + row[1])
                    nmr_param_oa = OntologyAnnotation(
                        term="standard concentration")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:SPECTROMETER FREQUENCY') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "spectrometer frequency: " + row[1])
                    nmr_param_oa = OntologyAnnotation(
                        term="spectrometer frequency")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:NMR_PROBE') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "nmr probe: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="nmr probe")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:NMR_SOLVENT') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "nmr solvent: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="nmr solvent")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:NMR_TUBE_SIZE') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "nmr tube size: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="nmr tube size")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:SHIMMING_METHOD') != -1:
                    # protocol_parameters["nmr spectroscopy"].append
                    # ("shimming method: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="shimming method")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:NMR_PULSE_SEQUENCE') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "nmr pulse sequence: " + row[1])
                    nmr_param_oa = OntologyAnnotation(
                        term="nmr pulse sequence")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:WATER_SUPPRESSION') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "water suppression: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="water suppression")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:PULSE_WIDTH') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "nmr pulse width: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="nmr pulse width")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:POWER_LEVEL') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "nmr power level: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="nmr power level")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:RECEIVER_GAIN') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "nmr receiver gain: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="nmr receiver gain")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:PRESATURATION_POWER_LEVEL') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "nmr presaturation power level: " + row[1])
                    nmr_param_oa = OntologyAnnotation(
                        term="nmr presaturation power level")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:CHEMICAL_SHIFT REFERENCE COMPOUND') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "chemical shift reference compound: " + row[1])
                    nmr_param_oa = OntologyAnnotation(
                        term="chemical shift reference compound")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:NUMBER_OF_SCANS') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "number of scans: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="number of scans")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:DUMMY_SCANS') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "dummy scans: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="dummy scans")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:ACQUISITION_TIME') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "acquisition time: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="acquisition time")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:RELAXATION_DELAY') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "relaxation delay: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="relaxation delay")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:SPECTRAL_WIDTH') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "spectral width: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="spectral width")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:LINE_BROADENING') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "line broadening: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="line broadering")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:ZERO_FILLING') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "zero filling: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="zero filling")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:APODIZATION') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "apodization: " + row[1])
                    nmr_param_oa = OntologyAnnotation(term="apodization")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('NMR:BASELINE_CORRECTION_METHOD') != -1:
                    # protocol_parameters["nmr spectroscopy"].append(
                    # "baseline correction method: " + row[1])
                    nmr_param_oa = OntologyAnnotation(
                        term="baseline correction method")
                    nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
                    nmr_protocol_params.append(nmr_param)

                if row[0].find('CO:COLLECTION_PROTOCOL_SUMMARY') != -1:
                    collect_protocol = collect_protocol + " " + row[1]

                if row[0].find('CO:COLLECTION_PROTOCOL_FILENAME') != -1:
                    protocol_files[3] = row[1]

                # if row[0].find('TR:PLANT_GROWTH_SUPPORT') != -1:
                #     plant_treatment_prtcl = \
                #         plant_treatment_prtcl + " " + row[1]

                if row[0].find('TR:PLANT_HARVEST_SUPPORT') != -1:
                    plant_harvest_method = plant_harvest_method + " " + row[1]

                if row[0].find('TR:PLANT_PLOT_DESIGN') != -1:
                    plant_plot_design = plant_plot_design + " " + row[1]

                if row[0].find('TR:PLANT_LIGHT_PERIOD') != -1:
                    plant_light_period = plant_light_period + " " + row[1]

                if row[0].find('TR:PLANT_TEMP') != -1:
                    plant_temperature = plant_temperature + " " + row[1]

                if row[0].find('TR:PLANT_NUTRITIONAL_REGIME') != -1:
                    plant_nutriregime = plant_nutriregime + " " + row[1]

                if row[0].find('TR:PLANT_METAB_QUENCH_METHOD') != -1:
                    quenching_method = quenching_method + " " + row[1]

                if row[0].find('SP:SAMPLEPREP_SUMMARY') != -1:
                    sampleprep_protocol = sampleprep_protocol + " " + row[1]

                if row[0].find('TR:TREATMENT_PROTOCOL_FILENAME') != -1:
                    protocol_files[1] = row[1]

                if row[0].find('SP:SAMPLEPREP_PROTOCOL_FILENAME') != -1:
                    protocol_files[2] = row[1]

            oa_st_design = OntologyAnnotation(term=study_design)
            # print(oa_st_design.name)

            study_filename = "s_" + studyid + ".txt"

            study1.title = study_title
            study1.description = str(study_desc)
            study1.submission_date = study_submittedon
            study1.public_release_date = study_createdon
            study1.design_descriptors = [oa_st_design]
            study1.filename = str(study_filename)

            for element in study_assays_dict["assays"]:
                # print("in study_assays_dict: ",element)
                if element["techtype"] == "mass spectrometry":
                    print("detected technology type:", element["techtype"])
                    orefTT = OntologySource(
                        name="OBI",
                        description="Ontology for Biomedical Investigation")
                    oaTT = OntologyAnnotation(
                        term="metabolite profiling",
                        term_accession="", term_source=orefTT)
                    orefMT = OntologySource(
                        name="OBI",
                        description="Ontology for Biomedical Investigation")
                    oaMT = OntologyAnnotation(
                        term="mass spectrometry",
                        term_accession="", term_source=orefMT)

                    this_assay_file = "a_" + str(studyid) + "_" + \
                                      str(element["analysis_id"]) + ".txt"
                    # print("this assay_file:", this_assay_file)
                    this_assay = Assay(measurement_type=oaTT,
                                       technology_type=oaMT,
                                       filename=this_assay_file)
                    study1.assays.append(this_assay)

                    downLoadURI = "http://www.metabolomicsworkbench.org/" \
                                  "data/study_textformat_view.php?STUDY_ID=" \
                                  + studyid \
                                  + "&ANALYSIS_ID="

                    downLoadURI = downLoadURI + element["analysis_id"] \
                        + "&MODE=d"

                    create_raw_data_files(
                        outputdir, tt, downLoadURI, studyid,
                        element["analysis_id"])

                    generate_maf_file(outputdir,
                                      studyid, element["analysis_id"])

                    assay_records, assay_header, qt1, qt2 = \
                        create_ms_assay_records(downLoadURI,
                                                studyid,
                                                element["analysis_id"],
                                                study_factor_records)

                    write_assay(
                        outputdir, element["techtype"], studyid,
                        element["analysis_id"], assay_records, assay_header)

                elif element["techtype"] == "nmr spectroscopy":
                    print("detected technology type:", element["techtype"])
                    orefTT = OntologySource(
                        name="OBI",
                        description="Ontology for Biomedical Investigation")
                    oaTT = OntologyAnnotation(
                        term="metabolite profiling",
                        term_accession="", term_source=orefTT)
                    orefMT = OntologySource(
                        name="OBI",
                        description="Ontology for Biomedical Investigation")
                    oaMT = OntologyAnnotation(
                        term="nmr spectroscopy",
                        term_accession="",
                        term_source=orefMT)
                    this_assay_file = "a_" + str(studyid) + "_" \
                                      + str(element["analysis_id"]) + ".txt"
                    # print("this NMR assay_file:", this_assay_file)
                    this_assay = Assay(measurement_type=oaTT,
                                       technology_type=oaMT,
                                       filename=this_assay_file)
                    study1.assays.append(this_assay)
                    # print("is it here?", study1.name)

                    downLoadURI = "http://www.metabolomicsworkbench.org/" \
                                  "data/study_textformat_view.php?STUDY_ID=" +\
                                  studyid + "&ANALYSIS_ID="

                    downLoadURI = \
                        downLoadURI + element["analysis_id"] + "&MODE=d"
                    print("invoking create_raw_data_method for NMR data now\n")
                    create_raw_data_files(
                        outputdir, tt, downLoadURI, studyid,
                        element["analysis_id"])

                    generate_maf_file(
                        outputdir, studyid, element["analysis_id"])

                    assay_records, assay_header, qt1, qt2 = \
                        create_nmr_assay_records(downLoadURI,
                                                 studyid,
                                                 element["analysis_id"],
                                                 study_factor_records)

                    write_assay(
                        outputdir, element["techtype"], studyid,
                        element["analysis_id"], assay_records, assay_header)

            chromatography_param_oa = \
                OntologyAnnotation(term="injection temperature")
            chromatography_param = \
                ProtocolParameter(parameter_name=chromatography_param_oa)
            chromatography_protocol_params.append(chromatography_param)
            chromatography_param_oa = OntologyAnnotation(term="flow rate")
            chromatography_param = \
                ProtocolParameter(parameter_name=chromatography_param_oa)
            chromatography_protocol_params.append(chromatography_param)
            chromatography_param_oa = OntologyAnnotation(term="solvent a")
            chromatography_param = \
                ProtocolParameter(parameter_name=chromatography_param_oa)
            chromatography_protocol_params.append(chromatography_param)
            chromatography_param_oa = OntologyAnnotation(term="solvent b")
            chromatography_param = \
                ProtocolParameter(parameter_name=chromatography_param_oa)
            chromatography_protocol_params.append(chromatography_param)
            chromatography_param_oa = \
                OntologyAnnotation(term="chromatography setting file")
            chromatography_param = \
                ProtocolParameter(parameter_name=chromatography_param_oa)
            chromatography_protocol_params.append(chromatography_param)

            ms_param_oa = OntologyAnnotation(term="ionization mode")
            ms_param = ProtocolParameter(parameter_name=ms_param_oa)
            ms_protocol_params.append(ms_param)
            ms_param_oa = OntologyAnnotation(term="injection mode")
            ms_param = ProtocolParameter(parameter_name=ms_param_oa)
            ms_protocol_params.append(ms_param)
            ms_param_oa = OntologyAnnotation(term="acquisition file")
            ms_param = ProtocolParameter(parameter_name=ms_param_oa)
            ms_protocol_params.append(ms_param)

            nmr_param_oa = OntologyAnnotation(term="nmr tube size")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(
                term="chemical shift reference compound")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(term="nmr solvent")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(term="zero filling")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(term="spectrometer frequency")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(term="relaxation delay")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(term="apodization")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(term="number of scans")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(
                term="baseline correction method")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(term="nmr solvent")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(term="nmr probe")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(term="nmr type")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(term="shimming method")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(term="nmr instrument")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)
            nmr_param_oa = OntologyAnnotation(term="pulse program")
            nmr_param = ProtocolParameter(parameter_name=nmr_param_oa)
            nmr_protocol_params.append(nmr_param)

            ident_param_oa = OntologyAnnotation(term="analysis file")
            ident_param = ProtocolParameter(parameter_name=ident_param_oa)
            ident_protocol_params.append(ident_param)
            ident_param_oa = OntologyAnnotation(term="software")
            ident_param = ProtocolParameter(parameter_name=ident_param_oa)
            ident_protocol_params.append(ident_param)

            oa_coll_prot = OntologyAnnotation(term="material collection")
            collection_protocol = Protocol(
                name="sample collection protocol",
                description=collect_protocol,
                uri=protocol_files[3],
                protocol_type=oa_coll_prot)
            study1.protocols.append(collection_protocol)

            oa_sample_prot = OntologyAnnotation(term="material preparation")
            sample_protocol = Protocol(
                name="metabolite extraction protocol",
                description=sampleprep_protocol,
                uri=protocol_files[2],
                protocol_type=oa_sample_prot)
            study1.protocols.append(sample_protocol)

            oa_treat = OntologyAnnotation(term="material processing")
            treatment_protocol = Protocol(name="treatment protocol",
                                          description=treat_protocol,
                                          uri=protocol_files[1],
                                          protocol_type=oa_treat)
            study1.protocols.append(treatment_protocol)

            oa_chromato_prot = OntologyAnnotation(term="chromatography")
            chromato_protocol = Protocol(
                name="chromatography protocol",
                description=chromatography_protocol,
                parameters=chromatography_protocol_params,
                uri="", protocol_type=oa_chromato_prot)
            study1.protocols.append(chromato_protocol)

            oa_ms_prot = OntologyAnnotation(term="mass spectrometry")
            ms_protocol = Protocol(name="mass spectrometry protocol",
                                   description=mass_spec_protocol,
                                   parameters=ms_protocol_params, uri="",
                                   protocol_type=oa_ms_prot)
            study1.protocols.append(ms_protocol)

            oa_nmr_prot = OntologyAnnotation(term="nmr spectroscopy")
            nmr_protocol = Protocol(
                name="nuclear magnetic resonance spectroscopy protocol",
                description=nmrspc_protocol,
                parameters=nmr_protocol_params,
                protocol_type=oa_nmr_prot)
            study1.protocols.append(nmr_protocol)

            oa_ident_prot = OntologyAnnotation(term="identification")
            ident_protocol = Protocol(name="identification protocol",
                                      description="ridiculous",
                                      protocol_type=oa_ident_prot,
                                      parameters=ident_protocol_params)
            study1.protocols.append(ident_protocol)

            study1.protocols.append(Protocol(
                name="annotation protocol",
                protocol_type=OntologyAnnotation(term="annotation")))

            publication = Publication(
                pubmed_id='12314444', status=OntologyAnnotation())
            study1.publications.append(publication)

            person1 = Person(first_name=study_person_firstname,
                             last_name=study_person_lastname,
                             email=study_person_email,
                             address=study_person_address,
                             affiliation=study_person_affiliation,
                             fax=study_person_fax,
                             comments=list())
            person1.comments.append(Comment(name="Grant Information",
                                            value=study_funder))
            study1.contacts.append(person1)

            # Building Investigation Study Factor Section:
            # factor_keys = study_factors.keys()
            for key in study_factors.keys():
                oref = OntologySource(
                    name="OBI",
                    description="Ontology for Biomedical Investigation")
                oa = OntologyAnnotation(
                    term=key, term_accession="", term_source=oref)
                study1.factors.append(StudyFactor(name=key, factor_type=oa))

            protocol_descriptions[0] = collect_protocol
            protocol_descriptions[1] = treat_protocol
            protocol_descriptions[2] = sampleprep_protocol

            studyfileheader = basestudysamplerecordheader + fv_record_header

            # ATTEMPTING TO WRITE STUDY FILES:
            write_study_file(outputdir, studyid, studyfileheader,
                             study_factor_records)

            # ATTEMPTING TO WRITE INVESTIGATION FILE:
            try:
                print("writing 'investigation information' to file...")
                print(isatab.dumps(investigation))
                isatab.dump(investigation, outputpath)
            except IOError:
                print("Error: in main() method can\'t open file or write data")
            # else:
            #  print("doh, something went wrong while writing investigation
            # but don't know why!: from main method")

            # ATTEMPTING TO DOWNLOAD THE CORRESPONDING DATA ARCHIVE
            # FROM MW ANONYMOUS FTP:
            if dl_option == 'yes':
                get_archived_file(studyid)
            elif dl_option == 'no':
                print('user elected not to dowload raw data')
            else:
                print('user input not recognized')

    except Exception as e:
        logging.exception(e)
        print("Error: in main() method something went wrong")
        print("conversion failed\n")
        conversion_success = False

    return conversion_success, studyid, validate_option



# -*- coding: utf-8 -*-
"""Functions for importing SRA from various sources"""
from __future__ import absolute_import
import functools
import logging
import os
import pdb
import re
import subprocess
import tempfile
import warnings
from io import BytesIO
from shutil import rmtree
from zipfile import ZipFile


def deprecated(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn(
            'Call to deprecated function {f}.'.format(f=func.__name__),
            category=DeprecationWarning, stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)
        return func(*args, **kwargs)

    return new_func


__author__ = 'massi'

DESTINATION_DIR = 'output'
DEFAULT_SAXON_EXECUTABLE = os.path.join(
    os.path.dirname(
        os.path.abspath(__file__)), 'resources', 'saxon9', 'saxon9he.jar')

SRA_DIR = os.path.join(os.path.dirname(__file__), 'resources', 'sra')

INPUT_FILE = os.path.join(SRA_DIR, 'blank.xml')

SUBMISSION_XSL_FILE = os.path.join(
    SRA_DIR, 'sra-submission-embl-online2isatab-txt.xsl')

STUDY_XSL_FILE = os.path.join(SRA_DIR, 'sra-study-embl-online2isatab.xsl')


log = logging.getLogger('isatools')

# REGEXES
_RX_ACCESSION_VALIDATION = re.compile(r"^(ERA|SRA|ERP|SRP)([0-9]+)$")


def zipdir(path, zip_file):
    """utility function to zip a whole directory"""
    # zip_file is zipfile handle
    for root, dirs, files in os.walk(path):
        for file in files:
            zip_file.write(os.path.join(root, file))


def format_acc_numbers(sra_acc_numbers):

    sra_acc_numbers = sra_acc_numbers.split(',') \
        if isinstance(sra_acc_numbers, str) else sra_acc_numbers

    # filter and clean the elements in the input array to match valid SRA types
    sra_acc_numbers = [elem.strip().upper() for elem in sra_acc_numbers]
    return [elem for elem in sra_acc_numbers
            if _RX_ACCESSION_VALIDATION.match(elem)]


@deprecated
def create_isatab_xslt(sra_acc_numbers, saxon_jar_path=None):
    """
    THIS METHOD IS DEPRECATED. USE sra_to_isatab_batch_convert INSTEAD.
    Given one or more SRA accession numbers (either to studies or
    submisssions),
    retrieve the files from the European Nucleotide Archive (ENA) server
    and convert them to ISA-tab using an XSL 2.0 transformation.
    The XSLT is invoked using an executable script.

    Notice: this method depend on SAXON XSLT Processor

    Parameters:
         :param sra_acc_numbers (str or list) - if a string must contain only
            comma separated valid SRA accession numbers
         :param saxon_jar_path str - if provided, must be a valid path
            pointing to SAXON Java JAR file. Otherwise, the
            default Saxon HE JAR will be used, if installed


    Returns:
        :returns zipfile.ZipFile if at least one of the SRA instances has been
            successfully converted
        :returns the output error otherwise (should this me modified?)

    """
    raise DeprecationWarning(
        'create_isatab_xslt() is deprecated, please use '
        'sra_to_isatab_batch_convert() instead')

    # cmd_map = dict(posix='batch_sra2isatab.sh', nt=None, java=None, ce=None)

    # formatted_sra_acc_numbers = format_acc_numbers(sra_acc_numbers)

    # convert the list back to a comma-separated string to be fed to the script
    # sra_acc_numbers_str = ",".join(formatted_sra_acc_numbers)

    # cmd_path = os.path.join(
    #     os.path.dirname(__file__), 'isa_line_commands', 'bin',
    #     cmd_map[os.name])
    pdb.set_trace()

    try:
        # res = subprocess.check_output([cmd_path, sra_acc_numbers_str])
        # put all files within a zip file and return the file handler
        file_like_obj = BytesIO()

        # return ZIP file containing all the created ISA-tab documents
        with ZipFile(file_like_obj, 'w') as zip_file:
            zipdir(DESTINATION_DIR, zip_file)
            return zip_file

    except subprocess.CalledProcessError as err:
        logging.error(
            'isatools.convert.sra2isatab: CalledProcessError caught ',
            err.returncode)
        return err.output


def sra_to_isatab_batch_convert(
        sra_acc_numbers, saxon_jar_path=DEFAULT_SAXON_EXECUTABLE):
    """
    Given one or more SRA accession numbers (either to studies or
    submisssions),
    retrieve the files from the European Nucleotide Archive (ENA) server
    and convert them to ISA-tab using an XSL 2.0 transformation.
    The XSLT is invoked using an executable script.

    Notice: this method depends on The Saxon XSLT and XQuery Processor from
    Saxonica Limited (http://www.saxonica.com)

    Parameters:
         :param sra_acc_numbers (str or list) - if a string must contain only
            comma separated valid SRA accession numbers
         :param saxon_jar_path str - if provided, must be a valid path
            pointing to SAXON Java JAR file. Otherwise, the default Saxon HE
            JAR will be used, if installed


    Returns:zp
        :returns io.BytesIO if at least one of the SRA instances has been
        successfully converted (NOTE: should I return StringIO instead?)

    """
    res = None
    destination_dir = None

    log.info('This function utilises The Saxon XSLT and XQuery Processor from '
             'Saxonica Limited (http://www.saxonica.com)')

    try:
        dir_name = tempfile.mkdtemp()
        formatted_sra_acc_numbers = format_acc_numbers(sra_acc_numbers)
        buffer = BytesIO()

        destination_dir = os.path.abspath(dir_name)
        log.info('Destination dir is: ' + destination_dir)

        for acc_number in formatted_sra_acc_numbers:
            try:

                if acc_number.startswith('SRA') or \
                        acc_number.startswith('ERA'):
                    res = subprocess.call(
                        ['java', '-jar', saxon_jar_path, INPUT_FILE,
                         STUDY_XSL_FILE, 'acc-number='+acc_number,
                         'outputdir='+destination_dir])

                elif acc_number.startswith('SRP') or \
                        acc_number.startswith('ERP'):
                    res = subprocess.call(
                        ['java', '-jar', saxon_jar_path, INPUT_FILE,
                         SUBMISSION_XSL_FILE, 'acc-number='+acc_number,
                         'outputdir='+destination_dir])

                log.info('Subprocess Saxon exited with code: %d', res)

                # post-process concatenation of a_ files written out
                output_folder = os.path.join(dir_name, acc_number)
                a_files = [
                    f for f in os.listdir(output_folder) if f.startswith('a_')]
                if len(a_files) > 1:
                    import pandas as pd
                    df_list = list()
                    for a_file in a_files:
                        df = pd.DataFrame()
                        a_path = os.path.join(output_folder, a_file)
                        df_list.append(df.from_csv(a_path, sep='\t'))
                        os.remove(a_path)
                    merged_a_file = pd.concat(df_list)
                    merged_a_file.to_csv(
                        os.path.join(dir_name, acc_number,
                                     'a_{}.txt'.format(acc_number)), sep='\t')

                with ZipFile(buffer, 'w') as zip_file:
                    # use relative dir_name to avoid absolute path on
                    # file names
                    zipdir(dir_name, zip_file)

            except subprocess.CalledProcessError as err:
                log.error(
                    'isatools.convert.sra2isatab: CalledProcessError caught ',
                    err.returncode)
        # clean up the target directory after the ZIP file has been closed
        # rmtree(destination_dir)

        buffer.seek(0)
    finally:
        log.debug('Removing dir' + destination_dir)
        rmtree(destination_dir)
    return buffer



# -*- coding: utf-8 -*
"""Convert mzML to ISA-Tab"""
import logging
import os
from mzml2isa import __version__

from mzml2isa.parsing import convert as mzml_convert

from isatools import isatab


logger = logging.getLogger('isatools')
logger.setLevel(logging.INFO)


def convert(mzml_folder, out_folder, study_id, validate_output=False):
    """ Converter for MZML to ISA-Tab.
    :param mzml_folder: Path to folder containing mzml files.
    :param out_folder: Path to output folder to write ISA-Tab to.
    :param study_id: A study identifier.
    :param validate_output: Flag to indicate whether to validate the generated
                            ISA-Tab
    """
    logger.info("mzML2ISA package version: {}".format(__version__))
    print("mzML2ISA package version: {}".format(__version__))
    print(logger)

    if not os.path.exists(mzml_folder):
        raise FileNotFoundError("Could not find input mzml folder")
    mzml_convert(mzml_folder, out_folder, study_id)
    if validate_output and os.path.exists(out_folder):
        with open(os.path.join(out_folder, 'i_Investigation.txt'),
                  'r', encoding='utf-8') as i_fp:
            return isatab.validate(i_fp)



# ERROR MESSAGES: PROTOCOL NODE
PARAMETER_VALUES_ERROR = 'The \'parameter_values\' property must be an iterable of isatools.model.ParameterValue ' \
                         'objects. {0} was supplied.'
REPLICATES_ERROR = 'Replicates must be a positive integer. {0} was supplied.'
PARAMETERS_CANNOT_BE_SET_ERROR = 'The \'parameters\' property cannot be set directly. Set parameter_values instead.'
COMPONENTS_CANNOT_BE_SET_ERROR = 'The \'components\' property cannot be set.'

# ERROR MESSAGES: PRODUCT NODE
NOT_ALLOWED_TYPE_ERROR = 'The provided ProductNode is not one of the allowed values: {0}'
PRODUCT_NODE_NAME_ERROR = 'ProductNode name must be a string, {0} supplied of type {1}'
SIZE_ERROR = 'ProductNode size must be a natural number, i.e integer >= 0'
CHARACTERISTIC_TYPE_ERROR = 'A characteristic must be either a string or a Characteristic, {0} supplied'
PRODUCT_NODE_EXTENSION_ERROR = 'ProductNode extension must be either a string or an OntologyAnnotation.'

# ERROR MESSAGES: QC SAMPLE (QUALITY CONTROL)
QC_SAMPLE_TYPE_ERROR = 'qc_sample_type must be one of {0}'

# ERROR MESSAGES: QUALITY CONTROL
PRE_BATCH_ATTRIBUTE_ERROR = 'Pre-batch must be an instance of ProductNode'
POST_BATCH_ATTRIBUTE_ERROR = 'Post-batch must be an instance of ProductNode'
INTERSPERSED_SAMPLE_TYPE_NODE_ERROR = 'Interspersed sample type must be an instance of ProductNode'
INTERSPERSED_SAMPLE_TYPE_INTERVAL_TYPE_ERROR = 'Sample type interval must be a positive integer'
INTERSPERSED_SAMPLE_TYPE_INTERVAL_VALUE_ERROR = 'Sample type interval must be a positive integer'

# ERROR MESSAGES: ASSAY GRAPH
INVALID_NODE_ERROR = 'Node must be instance of isatools.create.models.SequenceNode. {0} provided'
# INVALID_LINK_ERROR = "The link to be added is not valid. Link that can be created are ProductNode->ProtocolNode
# or ProtocolNode->ProductNode."
INVALID_LINK_ERROR = 'ProductNode->ProductNode links are not allowed in an assay workflow.'
INVALID_MEASUREMENT_TYPE_ERROR = '{0} is an invalid value for measurement_type. ' \
                                 'Please provide an OntologyAnnotation or string.'
INVALID_TECHNOLOGY_TYPE_ERROR = '{0} is an invalid value for technology_type. ' \
                                'Please provide an OntologyAnnotation or string.'
MISSING_NODE_ERROR = "Start or target node have not been added to the AssayGraph yet"
NODE_ALREADY_PRESENT = "The node {0.id} is already present in the AssayGraph"
QUALITY_CONTROL_ERROR = "The 'quality_control' must be a valid QualityControl object. {0} was supplied instead."

# ERROR MESSAGES: SAMPLE AND ASSAYPLAN
ASSAY_PLAN_NAME_ERROR = 'The attribute \'name\' must be a string. {0} provided'
MISSING_SAMPLE_IN_PLAN = 'ProductNode is missing from the sample_plan'
MISSING_ASSAY_IN_PLAN = 'AsssayGraph is missing from the assay_plan'

# ERROR MESSAGES: STUDY ARM
SCREEN_ERROR_MESSAGE = 'A SCREEN cell can only be inserted into an empty arm_map.'
RUN_IN_ERROR_MESSAGE = 'A RUN-IN cell can only be inserted into an arm_map containing a SCREEN.'
WASHOUT_ERROR_MESSAGE = 'A WASHOUT cell cannot be put next to a cell ending with a WASHOUT.'
COMPLETE_ARM_ERROR_MESSAGE = 'StudyArm complete. No more cells can be added after a FOLLOW-UP cell.'
FOLLOW_UP_ERROR_MESSAGE = 'A FOLLOW-UP cell cannot be put next to a SCREEN or a RUN-IN cell.'
FOLLOW_UP_EMPTY_ARM_ERROR_MESSAGE = 'A FOLLOW-UP cell cannot be put into an empty StudyArm.'
ARM_MAP_ASSIGNMENT_ERROR = 'arm_map must be an OrderedDict'
SOURCE_TYPE_ERROR = 'The source_type property must be either a string or a Characteristic. {0} was supplied.'

# ERROR MESSAGES: STUDY DESIGN
NAME_PROPERTY_ASSIGNMENT_ERROR = 'The value assigned to \'name\' must be a string'
DESIGN_TYPE_PROPERTY_ASSIGNMENT_ERROR = 'The value assigned to \'design_type\' must be a string or OntologyAnnotation'
DESCRIPTION_PROPERTY_ASSIGNMENT_ERROR = 'The value assigned to \'description\' must be text (i.e. string)'
STUDY_ARM_PROPERTY_ASSIGNMENT_ERROR = 'The value assigned to \'study_arms\' must be an iterable'
ADD_STUDY_ARM_PARAMETER_TYPE_ERROR = 'Not a valid study arm'
ADD_STUDY_ARM_NAME_ALREADY_PRESENT_ERROR = 'A StudyArm with the same name is already present in the StudyDesign'
GET_EPOCH_INDEX_OUT_OR_BOUND_ERROR = 'The Epoch you asked for is out of the bounds of the StudyDesign.'

# ERROR MESSAGES: STUDY DESIGN FACTORY
TREATMENT_MAP_ERROR = 'treatment_map must be a list containing tuples ' \
                      'with (Treatment, StudyAssayPlan) pairs.'
GROUP_SIZES_ERROR = 'no group sizes have been provided. Group size(s) must be provided either as an integer or as' \
                    'a tuple or list of integers'
GROUP_SIZE_ERROR = 'no group_size have been provided. Group size must be provided as an integer.'



# -*- coding: utf-8 -*-
"""Main entry point for running from the command-line.

This module provides the main entry point for running ISA-API functions from
the command-line. Commands must be run with a Python 3 interpreter.

Todo:
    * Implement hooks to all features of the ISA-API.
"""
import argparse
import json
import os
import sys


def main(argv=None):
    """Run **isatools** from the command line

    Args:
        argv (list, optional): the list of arguments to run isatools
        functions.

    Returns:
        None: Does not explicitly return anything apart from having the
        desired effect of the function that was called (usually output to
        file).
    """
    p = argparse.ArgumentParser(
        prog=__name__, formatter_class=argparse.RawDescriptionHelpFormatter,
        description='''Create, convert, and manipulate ISA-formatted
        metadata''', usage='isatools -c COMMAND [options]',)

    p.add_argument('-c', dest='cmd', help='isatools API command to run',
                   required=True,
                   choices=['isatab2json', 'json2isatab', 'sampletab2isatab',
                            'sampletab2json'])
    p.add_argument('-i', dest='in_path',
                   help='in  (files or directory will be read from here)',
                   required=True)
    p.add_argument('-o', dest='out_path',
                   help='out (file will be written out here or written to '
                        'directory if ISA-Tab archive out)', required=True)
    p.add_argument(
        '--version', action='version', version='isatools {}'.format(
            "0.10"))
    p.add_argument('-v', dest='verbose', help="show more output",
                   action='store_true', default=False)

    args = p.parse_args(argv or sys.argv[1:])

    if args.verbose:
        print("{} input: {}".format(os.linesep, args.in_path))
        print("output: {}".format(args.out_path))

    if args.cmd == 'isatab2json':
        from isatools.convert import isatab2json
        J = isatab2json.convert(args.in_path)
        with open(args.out_path, 'w') as out_fp:
            json.dump(J, out_fp)

    elif args.cmd == 'json2isatab':
        from isatools.convert import json2isatab
        with open(args.in_path) as in_fp:
            json2isatab.convert(in_fp, args.out_path)

    elif args.cmd == 'sampletab2isatab':
        from isatools.convert import sampletab2isatab
        with open(args.in_path) as in_fp:
            sampletab2isatab.convert(in_fp, args.out_path)

    elif args.cmd == 'sampletab2json':
        from isatools.convert import sampletab2json
        with open(args.in_path) as in_fp:
            with open(args.out_path, 'w') as out_fp:
                sampletab2json.convert(in_fp, out_fp)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*
""" ISA-specific exceptions """


class AttributeError(AttributeError):
    """
    If attempting to set a ISA Python object attribute to an invalid value.
    """


class ISAModelTypeError(TypeError):
    """
    If attempting to use a ISA Python object of a wrong type.
    """


class ISAModelValueError(ValueError):
    """
    If passing a parameter of the right type but the wrong value to an ISA model method.
    """


class ISAModelIndexError(IndexError):
    """
    If an index is out of bounds
    """





# -*- coding: utf-8 -*-
"""Functions for connecting the Ontology Lookup Service.

This module connects to the European Bioinformatics Institute's OLS.
If you have problems with it, check that it's working at
http://www.ebi.ac.uk/ols/
"""
from __future__ import absolute_import
import json
import logging
from urllib.request import urlopen

from isatools.model import OntologyAnnotation, OntologySource


OLS_API_BASE_URI = "http://www.ebi.ac.uk/ols/api"
OLS_PAGINATION_SIZE = 500


log = logging.getLogger('isatools')


def get_ols_ontologies():
    """Returns a list of OntologySource objects according to what's in OLS"""
    ontologiesUri = OLS_API_BASE_URI + "/ontologies?size=" + str(
        OLS_PAGINATION_SIZE)
    log.debug(ontologiesUri)
    J = json.loads(urlopen(ontologiesUri).read().decode("utf-8"))
    ontology_sources = []
    for ontology_source_json in J["_embedded"]["ontologies"]:
        ontology_sources.append(OntologySource(
            name=ontology_source_json["ontologyId"],
            version=ontology_source_json["config"]["version"],
            description=ontology_source_json["config"]["title"],
            file=ontology_source_json['_links']['self']
        ))
    return ontology_sources


def get_ols_ontology(ontology_name):
    """Returns a single OntologySource objects according to what's in OLS"""
    ontologiesUri = OLS_API_BASE_URI + "/ontologies?size=" + str(
        OLS_PAGINATION_SIZE)
    log.debug(ontologiesUri)
    J = json.loads(urlopen(ontologiesUri).read().decode("utf-8"))
    ontology_sources = []
    for ontology_source_json in J["_embedded"]["ontologies"]:
        ontology_sources.append(OntologySource(
            name=ontology_source_json["ontologyId"],
            version=ontology_source_json["config"]["version"],
            description=ontology_source_json["config"]["title"],
            file=ontology_source_json['_links']['self']['href']
        ))
    hits = [o for o in ontology_sources if o.name == ontology_name]
    if len(hits) == 1:
        return hits[0]
    else:
        return None


def search_ols(term, ontology_source):
    """Returns a list of OntologyAnnotation objects according to what's
    returned by OLS search"""
    url = OLS_API_BASE_URI + "/search"
    if isinstance(ontology_source, str):
        os_search = ontology_source
    elif isinstance(ontology_source, OntologySource):
        os_search = ontology_source.name
    else:
        os_search = None
    query = "{0}&queryFields=label&ontology={1}&exact=True".format(
        term, os_search)
    url += '?q={}'.format(query)
    log.debug(url)
    import requests
    req = requests.get(url)
    J = json.loads(req.text)
    ontology_annotations = []
    for search_result_json in J["response"]["docs"]:
        ontology_annotations.append(
            OntologyAnnotation(
                term=search_result_json["label"],
                term_accession=search_result_json["iri"],
                term_source=ontology_source if isinstance(
                    ontology_source, OntologySource) else None
            ))
    return ontology_annotations



from isatools.model import OntologyAnnotation, OntologySource, FactorValue, Characteristic
from isatools.create.model import StudyDesign, NonTreatment, Treatment, StudyCell, StudyArm, SampleAndAssayPlan
from isatools.create.constants import (
    SCREEN,
    INTERVENTIONS,
    BASE_FACTORS,
    SAMPLE,
    ORGANISM_PART,
    DEFAULT_SOURCE_TYPE,
    DATA_FILE,
    DEFAULT_EXTENSION
)
from collections import OrderedDict

AGENT = 'agent'

EVENT_TYPE_SAMPLING = 'sampling'
EVENT_TYPE_ASSAY = 'assay'


def _map_ontology_annotation(annotation, expand_strings=False):
    """
    converts an input annotation into an OntologyAnnotation. If the input is a string it is kept as it is and not
    cast as an OntologyAnnotation
    :param annotation: str/dict
    :return: str/OntologyAnnotation
    """
    if isinstance(annotation, dict):
        res = OntologyAnnotation(
            term=annotation['term']
        )
        if annotation.get('iri', None):
            res.term_accession = annotation['iri']
        source = annotation.get('source', None)
        if source:
            if isinstance(source, str):
                res.term_source = OntologySource(name=source)
            elif isinstance(source, dict):
                res.term_source = OntologySource(**source)
        return res
    else:
        return OntologyAnnotation(term=annotation) if expand_strings else annotation


def _reverse_map_ontology_annotation(onto_annotation, compress_strings=False):
    """
    converts an OntologyAnnotation into its serializable form (i.e. a dict with string keys)
    no comments or ids are serialized by this special serializer
    if the input is a string the string itself is returned.
    :param onto_annotation: str or OntologyAnnotation
    :param compress_strings Bool - if true compress to strings onto_annotation objects missing a term_accession
    :return: dict or string - to be serialized as JSON
    """
    if isinstance(onto_annotation, OntologyAnnotation):
        if not onto_annotation.term_accession and compress_strings:
            return onto_annotation.term
        res = dict(term=onto_annotation.term, iri=onto_annotation.term_accession or None)
        if onto_annotation.term_source:
            if isinstance(onto_annotation.term_source, str):
                res['source'] = onto_annotation.term_source
            elif isinstance(onto_annotation.term_source, OntologySource):
                res['source'] = dict(
                    name=onto_annotation.term_source.name,
                    file=onto_annotation.term_source.file,
                    version=onto_annotation.term_source.version,
                    description=onto_annotation.term_source.description
                )
        else:
            res['source'] = None
        return res
    else:
        return onto_annotation


def assay_template_to_ordered_dict(assay_template):
    """
    deserializes a JSON plain dictionary into an OrderedDictionary to be consumed by
        isatools.create.models.AssayGraph.generate_assay_plan_from_dict() to generate a full AssayGraph
    :param assay_template: dict
    :return: OrderedDict.
    """
    res = OrderedDict()
    res['measurement_type'] = _map_ontology_annotation(assay_template['measurement_type'], expand_strings=True)
    res['technology_type'] = _map_ontology_annotation(assay_template['technology_type'], expand_strings=True)
    for name, nodes in assay_template['workflow']:
        prepared_nodes = None
        if isinstance(nodes, list):
            # "nodes" represent a list of ProductNodes
            prepared_nodes = [
                {
                    key: _map_ontology_annotation(
                        value, expand_strings=True if key in ['characteristics_category'] else False
                    ) for key, value in el.items()
                } for el in nodes
            ]
        if isinstance(nodes, dict):
            # "nodes" represent a ProtocolNode
            prepared_nodes = {}
            for candidate_param_name, param_values in nodes.items():
                # if it is a special key (e.g."#replicates") leave it alone
                if candidate_param_name[0] == '#' and not isinstance(param_values, list):
                    prepared_nodes[candidate_param_name] = param_values
                else:
                    # this is really a parameter name
                    param_name = _map_ontology_annotation(candidate_param_name, expand_strings=True)
                    prepared_nodes[param_name] = [
                        _map_ontology_annotation(param_value) for param_value in param_values
                    ]
        res[_map_ontology_annotation(name)] = prepared_nodes
    return res


def assay_ordered_dict_to_template(assay_ord_dict):
    """
    Serializes an OrderedDict compatible with isatools.create.models.AssayGraph.generate_assay_plan_from_dict() into a
        JSON representation that can be consumed by external applications (e.g. Datascriptor and other client
        applications)
    :param assay_ord_dict: OrderedDictionary
    :return: dict, can be directly serialized to JSON
    """
    res = dict()
    res['measurement_type'] = _reverse_map_ontology_annotation(assay_ord_dict['measurement_type'],
                                                               compress_strings=True)
    res['technology_type'] = _reverse_map_ontology_annotation(assay_ord_dict['technology_type'],
                                                              compress_strings=True)
    res['workflow'] = []
    for name, nodes in assay_ord_dict.items():
        if name in {'measurement_type', 'technology_type'}:
            continue
        if isinstance(nodes, dict):
            # "nodes" represent a ProtocolNode
            serialized_nodes = {}
            for node_prop, prop_values in nodes.items():
                if isinstance(prop_values, list):
                    serialized_nodes[_reverse_map_ontology_annotation(node_prop, compress_strings=True)] = [
                        _reverse_map_ontology_annotation(prop_value) for prop_value in prop_values
                    ]
                else:
                    serialized_nodes[node_prop] = prop_values
        elif isinstance(nodes, list):
            # "nodes" represent a list of ProductNodes
            serialized_nodes = [
                {
                    key: _reverse_map_ontology_annotation(
                        val, compress_strings=True if key in ['characteristics_category'] else False
                    ) for key, val in node.items()
                } for node in nodes
            ]
        else:
            serialized_nodes = {}
        res['workflow'].append([
            _reverse_map_ontology_annotation(name),
            serialized_nodes
        ])
    return res


def _generate_element(datascriptor_element_dict):
    """
    Generates elements (Treatement and NonTreatment) from their description as Datascriptor dicts
    :param datascriptor_element_dict: dict
    :return: isatools.create.models.Element
    """
    if 'agent' in datascriptor_element_dict:
        agent = FactorValue(
            factor_name=BASE_FACTORS[0],
            value=_map_ontology_annotation(datascriptor_element_dict['agent']),
            unit=_map_ontology_annotation(datascriptor_element_dict.get('agentUnit', None))
        )
        intensity = FactorValue(
            factor_name=BASE_FACTORS[1],
            value=_map_ontology_annotation(datascriptor_element_dict.get('intensity', None)),
            unit=_map_ontology_annotation(datascriptor_element_dict.get('intensityUnit', None))
        )
        duration = FactorValue(
            factor_name=BASE_FACTORS[2],
            value=datascriptor_element_dict.get('duration', 0),
            unit=_map_ontology_annotation(datascriptor_element_dict.get('durationUnit', 's'))
        )
        intervention_type = datascriptor_element_dict.get('interventionType', '')
        element = Treatment(
            element_type=_map_ontology_annotation(intervention_type) if intervention_type
            else INTERVENTIONS['UNSPECIFIED'],
            factor_values=[agent, intensity, duration]
        )
    else:
        element = NonTreatment(
            element_type=datascriptor_element_dict.get('name', SCREEN),
            duration_value=datascriptor_element_dict['duration'],
            duration_unit=_map_ontology_annotation(datascriptor_element_dict['durationUnit'])
        )
    return element


def _generate_sample_dict_from_config(datascriptor_sample_type_config, arm_name, epoch_no):
    return dict(
        node_type=SAMPLE,
        characteristics_category=_map_ontology_annotation(
            datascriptor_sample_type_config.get('characteristicCategory', ORGANISM_PART),
            expand_strings=True
        ),
        characteristics_value=_map_ontology_annotation(datascriptor_sample_type_config['sampleType']),
        size=datascriptor_sample_type_config['sampleTypeSizes'][arm_name][epoch_no],
        is_input_to_next_protocols=datascriptor_sample_type_config.get('isAssayInput', True)
    )


def _generate_characteristics_from_observational_factor(observational_factor_dict):
    category = _map_ontology_annotation(observational_factor_dict['name'], expand_strings=True)
    value = _map_ontology_annotation(
        observational_factor_dict['value'], expand_strings=True
    ) if observational_factor_dict['isQuantitative'] is False else observational_factor_dict['value']
    unit = _map_ontology_annotation(
        observational_factor_dict['unit'], expand_strings=True
    ) if observational_factor_dict['isQuantitative'] is True else None
    return Characteristic(category=category, value=value, unit=unit)


def generate_assay_ord_dict_from_config(datascriptor_assay_config, arm_name, epoch_no):
    res = OrderedDict()
    res['id'], res['name'] = datascriptor_assay_config['id'], datascriptor_assay_config['name']
    res['measurement_type'] = _map_ontology_annotation(
        datascriptor_assay_config['measurement_type'], expand_strings=True
    )
    res['technology_type'] = _map_ontology_annotation(
        datascriptor_assay_config['technology_type'], expand_strings=True
    )
    res['selected_sample_types'] = list(map(
        _map_ontology_annotation,
        datascriptor_assay_config['selectedSampleTypes'][arm_name][epoch_no]
    ))
    for name, node in datascriptor_assay_config['workflow']:
        prepared_nodes = None
        assert isinstance(node, dict)
        if '#replicates' in node:
            # this is a ProtocolNode
            prepared_nodes = {}
            for candidate_param_name, param in node.items():
                # if it is a special key (e.g."#replicates") leave it alone
                if candidate_param_name[0] == '#' and not isinstance(param, list):
                    prepared_nodes[candidate_param_name] = param['value']
                else:
                    if not param['values']:
                        raise ValueError('Missing values for Protocol Param {}'.format(
                            candidate_param_name['term'] if isinstance(
                                candidate_param_name, dict
                            ) else candidate_param_name
                        ))
                    # this is really a parameter name
                    param_name = _map_ontology_annotation(candidate_param_name, expand_strings=True)
                    prepared_nodes[param_name] = [
                        _map_ontology_annotation(param_value) for param_value in param['values']
                    ]
        elif 'node_type' in node:
            # this is a product node
            extension = node['extension']['value'] if 'extension' in node else DEFAULT_EXTENSION \
                if node['node_type'] == DATA_FILE else None
            if "characteristics_value" in node:
                if not node["characteristics_value"]['values']:
                    raise ValueError('Missing values for Characteristic {}'.format(
                        node['characteristics_category']['term'] if isinstance(
                            node['characteristics_category'], dict
                        ) else node['characteristics_category']
                    ))
                prepared_nodes = [
                    dict(
                        node_type=node['node_type'],
                        characteristics_category=_map_ontology_annotation(node['characteristics_category'],
                                                                          expand_strings=True),
                        characteristics_value=_map_ontology_annotation(value),
                        size=node.get('size', 1),
                        is_input_to_next_protocols=node['is_input_to_next_protocols']['value'],
                        extension=extension
                    ) for value in node["characteristics_value"]["values"]
                ]
            else:
                prepared_nodes = [dict(
                    node_type=node['node_type'],
                    size=node.get('size', 1),
                    is_input_to_next_protocols=node['is_input_to_next_protocols']['value'],
                    extension=extension
                )]

        res[_map_ontology_annotation(name)] = prepared_nodes
    return res


def generate_study_design(datascriptor_study_config):
    """
    This function takes a study design configuration as produced from the Datascriptor application
    and outputs a StudyDesign object
    :param datascriptor_study_config: dict
    :return: isatools.create.StudyDesign
    """
    study_design_config = datascriptor_study_config['design']
    arms = []
    for arm_ix, arm_dict in enumerate(study_design_config['arms']['selected']):
        arm_map = OrderedDict()
        for epoch_ix, epoch_dict in enumerate(arm_dict['epochs']):
            element_ids = epoch_dict.get('elements', [])
            elements = [
                _generate_element(element_dict) for element_dict in
                filter(
                    lambda el: el['id'] in element_ids,
                    study_design_config['elements']
                )
            ]
            cell_name = 'A{}E{}'.format(arm_ix, epoch_ix)
            cell = StudyCell(name=cell_name, elements=elements)
            sample_type_dicts = [
                _generate_sample_dict_from_config(
                    ds_sample_config, arm_dict['name'], epoch_ix
                ) for ds_sample_config in study_design_config['samplePlan']
                if ds_sample_config['selectedCells'][arm_dict['name']][epoch_ix] and
                ds_sample_config['sampleTypeSizes'][arm_dict['name']][epoch_ix]
            ]
            assay_ord_dicts = [
                generate_assay_ord_dict_from_config(
                    ds_assay_config, arm_dict['name'], epoch_ix
                ) for ds_assay_config in study_design_config['assayPlan']
                if ds_assay_config['selectedCells'][arm_dict['name']][epoch_ix] is True
            ]
            sa_plan_name = 'SAP_A{}E{}'.format(arm_ix, epoch_ix)
            # TODO this method will probably need some rework to bind a sample type to a specific assay plan
            sa_plan = SampleAndAssayPlan.from_sample_and_assay_plan_dict(
                sa_plan_name, sample_type_dicts, *assay_ord_dicts
            )
            arm_map[cell] = sa_plan
        source_type = Characteristic(
            category=DEFAULT_SOURCE_TYPE.category,
            value=_map_ontology_annotation(
                arm_dict.get('subjectType', None) or study_design_config.get('subjectType', None)
            )
        )
        arm = StudyArm(
            name=arm_dict['name'],
            # should we generate a Characteristic if subjectType is an OntologyAnnotation?
            source_type=source_type,
            source_characteristics=[
                _generate_characteristics_from_observational_factor(
                    obs_factor_dict
                ) for obs_factor_dict in arm_dict.get('observationalFactors', [])
            ],
            group_size=arm_dict.get('size', 10),
            arm_map=arm_map
        )
        arms.append(arm)
    return StudyDesign(
        identifier=datascriptor_study_config.get('_id', None),
        name=datascriptor_study_config['name'],
        description=datascriptor_study_config.get('description', None),
        design_type=_map_ontology_annotation(study_design_config['designType']),
        study_arms=arms
    )



#!/usr/bin/env python
import os
from setuptools import setup


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(f_name):
    return open(os.path.join(os.path.dirname(__file__), f_name)).read()


setup(
    name='isatools',
    version='0.13.0-rc.1',
    packages=['isatools',
              'isatools.convert',
              'isatools.create',
              'isatools.errors',
              'isatools.io',
              'isatools.net',
              'isatools.tests'
              ],
    package_data={'isatools': [
        'resources/schemas/cedar/*.json',
        'resources/schemas/isa_model_version_1_0_schemas/core/*.json',
        'resources/schemas/configs/*.json',
        'resources/schemas/configs/schemas/*.json',
        'resources/config/json/default/*.json',
        'resources/config/json/default/schemas/*.json',
        'resources/config/json/sra/*.json',
        'resources/config/json/sra/schemas/*.json',
        'resources/config/xml/*.xml',
        'resources/config/yaml/*.yml',
        'resources/sra_schemas/*.xsd',
        'resources/sra_templates/*.xml',
        'resources/tab_templates/*.txt',
        'net/resources/biocrates/*',
        'net/resources/sra/*.xsl',
        'net/resources/sra/*.xml',
        'resources/isatools.ini'],
        '': ['LICENSE.txt', 'README.md']},
    description='Metadata tracking tools help to manage an increasingly diverse set of life science, '
                'environmental and biomedical experiments',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='ISA Infrastructure Team',
    author_email='isatools@googlegroups.com',
    url='https://github.com/ISA-tools/isa-api',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    install_requires=[
        'numpy~=1.21.0',
        'jsonschema~=3.2.0',
        'pandas~=1.2.5',
        'networkx~=2.5.1',
        'lxml~=4.6.3',
        'requests~=2.25.1',
        'chardet~=4.0.0',
        'iso8601~=0.1.14',
        'jinja2~=3.0.1',
        'beautifulsoup4~=4.9.3',
        'mzml2isa==1.0.3',
        'biopython~=1.79',
        'progressbar2~=3.53.1',
        'deepdiff~=5.5.0',
        'PyYAML~=5.4.1'
    ],
    test_suite='tests'
)



# -*- coding: utf-8 -*-
"""Utilities for logging in the ISA-API."""
from __future__ import absolute_import
import logging
import os
from configparser import ConfigParser


show_pbars = False


def read(path):
    """Read a .ini config to set up some global defaults

    :param path: Path to logging settings ini file. Defaults to
    resources/isatools.ini
    :return: None
    """
    global log_level
    global show_pbars
    cparser = ConfigParser()
    cparser.read(path)
    log_level_ini = cparser.get('Logging', 'LogLevel')
    # print("log_level:", log_level_ini.lower())
    if log_level_ini.lower() == 'warning':
        log_level = logging.WARNING
    elif log_level_ini.lower() == 'debug':
        log_level = logging.DEBUG
    elif log_level_ini.lower() == 'error':
        log_level = logging.ERROR
    elif log_level_ini.lower() == 'critical':
        log_level = logging.CRITICAL
    else:
        log_level = logging.INFO
    log_format = "%(asctime)s [%(levelname)s]: " \
                 "%(filename)s(%(funcName)s:%(lineno)s) >> %(message)s"
    logging.basicConfig(format=log_format)
    set_level(log_level=log_level)

    showpbars_ini = cparser.get('Logging', 'showprogressbars')
    if showpbars_ini.lower() == 'yes':
        show_pbars = True
    else:
        show_pbars = False


def set_level(log_level):
    """Set the logging level

    :param log_level: Based on Python logging module
    :return: None
    """
    if log_level in (logging.NOTSET, logging.DEBUG, logging.INFO,
                     logging.WARNING, logging.ERROR, logging.CRITICAL):
        logging.getLogger('isatools').setLevel(log_level)


read(os.path.join(os.path.dirname(
    os.path.abspath(__file__)), 'resources', 'isatools.ini'))



# -*- coding: utf-8 -*-
"""isatools meta-module for importing various submodules.

This module allows users to import submodules without having to use fully
qualified import indices by exposing all submodules on the same level
entry point.

Example:

    Instead of importing the isatab2json converter directly::

        $ from isatools.convert import isatab2json

    We can import the same module without specifying the .convert package::

        $ from isatools import isatab2json
"""
from __future__ import absolute_import

from isatools.convert import (
    isatab2cedar as isatab2cedar_module,
    isatab2json as isatab2json_module,
    isatab2sampletab as isatab2sampletab_module,
    isatab2sra as isatab2sra_module,
    isatab2w4m as isatab2w4m_module,
    json2isatab as json2isatab_module,
    json2magetab as json2magetab_module,
    json2sampletab as json2sampletab_module,
    json2sra as json2sra_module,
    magetab2isatab as magetab2isatab_module,
    magetab2json as magetab2json_module,
    mzml2isa as mzml2isa_module,
    sampletab2isatab as sampletab2isatab_module,
    sampletab2json as sampletab2json_module,
)
from isatools.net import (
    biocrates2isatab as biocrates2isatab_module,
    mtbls as mtbls_module,
    mw2isa as mw2isa_module,
    ols as ols_module,
    pubmed as pubmed_module,
    sra2isatab as sra2isatab_module,
)


# isatools.convert packages
isatab2cedar = isatab2cedar_module
isatab2json = isatab2json_module
isatab2sampletab = isatab2sampletab_module
isatab2sra = isatab2sra_module
isatab2w4m = isatab2w4m_module
json2isatab = json2isatab_module
json2magetab = json2magetab_module
json2sampletab = json2sampletab_module
json2sra = json2sra_module
magetab2isatab = magetab2isatab_module
magetab2json = magetab2json_module
mzml2isa = mzml2isa_module
sampletab2isatab = sampletab2isatab_module
sampletab2json = sampletab2json_module

# isatools.net packages
biocrates2isatab = biocrates2isatab_module
mtbls = mtbls_module
mw2isa = mw2isa_module
ols = ols_module
pubmed = pubmed_module
sra2isatab = sra2isatab_module



# -*- coding: utf-8 -*
"""Convert ISA-JSON to SampleTab"""
import logging

from isatools import isajson, sampletab


log = logging.getLogger('isatools')


def convert(source_json_fp, target_fp):
    """ Converter for ISA-JSON to SampleTab.
    :param source_json_fp: File descriptor of input ISA JSON file
    :param target_fp: File descriptor to write output SampleTab to (must be
                      writeable)
    """
    log.info("loading isajson %s", source_json_fp.name)
    ISA = isajson.load(source_json_fp)
    log.info("dumping sampletab %s", target_fp.name)
    sampletab.dump(ISA, target_fp)
